# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from parameterized import parameterized

import mock

from twisted.internet import defer
from twisted.internet import task
from twisted.trial import unittest

from buildbot import config
from buildbot.changes import changes
from buildbot.changes import filter
from buildbot.process import properties
from buildbot.process.properties import Interpolate
from buildbot.schedulers import base
from buildbot.test import fakedb
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.util import scheduler
from buildbot.test.util.warnings import assertProducesWarning
from buildbot.warnings import DeprecatedApiWarning


class BaseScheduler(scheduler.SchedulerMixin, TestReactorMixin,
                    unittest.TestCase):

    OBJECTID = 19
    SCHEDULERID = 9
    exp_bsid_brids = (123, {'b': 456})

    def setUp(self):
        self.setup_test_reactor()
        self.setUpScheduler()

    def tearDown(self):
        self.tearDownScheduler()

    def makeScheduler(self, name='testsched', builderNames=None,
                      properties=None, codebases=None):
        if builderNames is None:
            builderNames = ['a', 'b']
        if properties is None:
            properties = {}
        if codebases is None:
            codebases = {'': {}}

        if isinstance(builderNames, list):
            dbBuilder = []
            builderid = 0
            for builderName in builderNames:
                builderid += 1
                dbBuilder.append(fakedb.Builder(id=builderid, name=builderName))

            self.master.db.insertTestData(dbBuilder)

        sched = self.attachScheduler(
            base.BaseScheduler(name=name, builderNames=builderNames,
                               properties=properties, codebases=codebases),
            self.OBJECTID, self.SCHEDULERID)
        self.master.data.updates.addBuildset = mock.Mock(
            name='data.addBuildset',
            side_effect=lambda *args, **kwargs:
            defer.succeed(self.exp_bsid_brids))

        return sched

    # tests

    def test_constructor_builderNames(self):
        with self.assertRaises(config.ConfigErrors):
            self.makeScheduler(builderNames='xxx')

    def test_constructor_builderNames_unicode(self):
        self.makeScheduler(builderNames=['a'])

    def test_constructor_builderNames_renderable(self):
        @properties.renderer
        def names(props):
            return ['a']
        self.makeScheduler(builderNames=names)

    def test_constructor_codebases_valid(self):
        codebases = {"codebase1":
                     {"repository": "", "branch": "", "revision": ""}}
        self.makeScheduler(codebases=codebases)

    def test_constructor_codebases_valid_list(self):
        codebases = ['codebase1']
        self.makeScheduler(codebases=codebases)

    def test_constructor_codebases_invalid(self):
        # scheduler only accepts codebases with at least repository set
        codebases = {"codebase1": {"dictionary": "", "that": "", "fails": ""}}
        with self.assertRaises(config.ConfigErrors):
            self.makeScheduler(codebases=codebases)

    @defer.inlineCallbacks
    def test_getCodebaseDict(self):
        sched = self.makeScheduler(
            codebases={'lib': {'repository': 'librepo'}})
        cbd = yield sched.getCodebaseDict('lib')
        self.assertEqual(cbd, {'repository': 'librepo'})

    @defer.inlineCallbacks
    def test_getCodebaseDict_constructedFromList(self):
        sched = self.makeScheduler(codebases=['lib', 'lib2'])
        cbd = yield sched.getCodebaseDict('lib')
        self.assertEqual(cbd, {})

    def test_getCodebaseDict_not_found(self):
        sched = self.makeScheduler(
            codebases={'lib': {'repository': 'librepo'}})
        return self.assertFailure(sched.getCodebaseDict('app'), KeyError)

    def test_listBuilderNames(self):
        sched = self.makeScheduler(builderNames=['x', 'y'])
        self.assertEqual(sched.listBuilderNames(), ['x', 'y'])

    @defer.inlineCallbacks
    def test_startConsumingChanges_fileIsImportant_check(self):
        sched = self.makeScheduler()
        try:
            yield sched.startConsumingChanges(fileIsImportant="maybe")
        except AssertionError:
            pass
        else:
            self.fail("didn't assert")

    @defer.inlineCallbacks
    def test_enabled_callback(self):
        sched = self.makeScheduler()
        expectedValue = not sched.enabled
        yield sched._enabledCallback(None, {'enabled': not sched.enabled})
        self.assertEqual(sched.enabled, expectedValue)
        expectedValue = not sched.enabled
        yield sched._enabledCallback(None, {'enabled': not sched.enabled})
        self.assertEqual(sched.enabled, expectedValue)

    @defer.inlineCallbacks
    def do_test_change_consumption(self, kwargs, expected_result, change_kwargs=None):
        if change_kwargs is None:
            change_kwargs = {}

        # (expected_result should be True (important), False (unimportant), or
        # None (ignore the change))
        sched = self.makeScheduler()
        sched.startService()
        self.addCleanup(sched.stopService)

        # set up a change message, a changedict, a change, and convince
        # getChange and fromChdict to convert one to the other
        msg = dict(changeid=12934)

        chdict = dict(changeid=12934, is_chdict=True)

        def getChange(changeid):
            assert changeid == 12934
            return defer.succeed(chdict)
        self.db.changes.getChange = getChange

        change = self.makeFakeChange(**change_kwargs)
        change.number = 12934

        def fromChdict(cls, master, chdict):
            assert chdict['changeid'] == 12934 and chdict['is_chdict']
            return defer.succeed(change)
        self.patch(changes.Change, 'fromChdict', classmethod(fromChdict))

        change_received = [None]

        def gotChange(got_change, got_important):
            # check that we got the expected change object
            self.assertIdentical(got_change, change)
            change_received[0] = got_important
            return defer.succeed(None)
        sched.gotChange = gotChange

        yield sched.startConsumingChanges(**kwargs)

        # check that it registered callbacks
        self.assertEqual(len(self.mq.qrefs), 2)

        qref = self.mq.qrefs[1]
        self.assertEqual(qref.filter, ('changes', None, 'new'))

        # invoke the callback with the change, and check the result
        qref.callback('change.12934.new', msg)
        self.assertEqual(change_received[0], expected_result)

    def test_change_consumption_defaults(self):
        # all changes are important by default
        return self.do_test_change_consumption(
            {},
            True)

    def test_change_consumption_fileIsImportant_True(self):
        return self.do_test_change_consumption(
            dict(fileIsImportant=lambda c: True),
            True)

    def test_change_consumption_fileIsImportant_False(self):
        return self.do_test_change_consumption(
            dict(fileIsImportant=lambda c: False),
            False)

    @defer.inlineCallbacks
    def test_change_consumption_fileIsImportant_exception(self):
        yield self.do_test_change_consumption(
            dict(fileIsImportant=lambda c: 1 / 0),
            None)

        self.assertEqual(1, len(self.flushLoggedErrors(ZeroDivisionError)))

    def test_change_consumption_change_filter_True(self):
        cf = mock.Mock()
        cf.filter_change = lambda c: True
        return self.do_test_change_consumption(
            dict(change_filter=cf),
            True)

    def test_change_consumption_change_filter_False(self):
        cf = mock.Mock()
        cf.filter_change = lambda c: False
        return self.do_test_change_consumption(
            dict(change_filter=cf),
            None)

    def test_change_consumption_change_filter_gerrit_ref_updates(self):
        cf = mock.Mock()
        cf.filter_change = lambda c: False
        return self.do_test_change_consumption(
            {'change_filter': cf},
            None,
            change_kwargs={'category': 'ref-updated', 'branch': 'master'})

    def test_change_consumption_change_filter_gerrit_ref_updates_with_refs(self):
        cf = mock.Mock()
        cf.filter_change = lambda c: False
        return self.do_test_change_consumption(
            {'change_filter': cf},
            None,
            change_kwargs={'category': 'ref-updated', 'branch': 'refs/changes/123'})

    @parameterized.expand(
        [
            (
                "branch",
                filter.ChangeFilter(branch='refs/heads/master'),
                True
            ),
            (
                "branch_re",
                filter.ChangeFilter(branch_re='refs/heads/master'),
                True
            ),
            (
                "branch_re_no_match",
                filter.ChangeFilter(branch_re='(refs/heads/other|master)'),
                None
            )
        ]
    )
    def test_change_consumption_refs_heads_branch_deprecated(self, name, cf, expected_result):
        with assertProducesWarning(DeprecatedApiWarning,
                                   "Change filters must not expect ref-updated events"):
            return self.do_test_change_consumption(
                {'change_filter': cf},
                expected_result,
                change_kwargs={'category': 'ref-updated', 'branch': 'master'}
            )

    def test_change_consumption_change_filter_gerrit_filters_branch_new(self):
        cf = filter.ChangeFilter(branch='master')
        return self.do_test_change_consumption(
            {'change_filter': cf},
            True,
            change_kwargs={'category': 'ref-updated', 'branch': 'master'})

    def test_change_consumption_change_filter_gerrit_filters_branch_new_not_match(self):
        cf = filter.ChangeFilter(branch='other')
        return self.do_test_change_consumption(
            {'change_filter': cf},
            None,
            change_kwargs={'category': 'ref-updated', 'branch': 'master'})

    def test_change_consumption_fileIsImportant_False_onlyImportant(self):
        return self.do_test_change_consumption(
            dict(fileIsImportant=lambda c: False, onlyImportant=True),
            None)

    def test_change_consumption_fileIsImportant_True_onlyImportant(self):
        return self.do_test_change_consumption(
            dict(fileIsImportant=lambda c: True, onlyImportant=True),
            True)

    @defer.inlineCallbacks
    def test_activation(self):
        sched = self.makeScheduler(name='n', builderNames=['a'])
        sched.clock = task.Clock()
        sched.activate = mock.Mock(return_value=defer.succeed(None))
        sched.deactivate = mock.Mock(return_value=defer.succeed(None))

        # set the schedulerid, and claim the scheduler on another master
        yield self.setSchedulerToMaster(self.OTHER_MASTER_ID)

        yield sched.startService()
        sched.clock.advance(sched.POLL_INTERVAL_SEC / 2)
        sched.clock.advance(sched.POLL_INTERVAL_SEC / 5)
        sched.clock.advance(sched.POLL_INTERVAL_SEC / 5)
        self.assertFalse(sched.activate.called)
        self.assertFalse(sched.deactivate.called)
        self.assertFalse(sched.isActive())
        # objectid is attached by the test helper
        self.assertEqual(sched.serviceid, self.SCHEDULERID)

        # clear that masterid
        yield sched.stopService()
        self.setSchedulerToMaster(None)
        yield sched.startService()
        sched.clock.advance(sched.POLL_INTERVAL_SEC)
        self.assertTrue(sched.activate.called)
        self.assertFalse(sched.deactivate.called)
        self.assertTrue(sched.isActive())

        # stop the service and see that deactivate is called
        yield sched.stopService()
        self.assertTrue(sched.activate.called)
        self.assertTrue(sched.deactivate.called)
        self.assertFalse(sched.isActive())

    def test_activation_claim_raises(self):
        sched = self.makeScheduler(name='n', builderNames=['a'])
        sched.clock = task.Clock()

        # set the schedulerid, and claim the scheduler on another master
        self.setSchedulerToMaster(RuntimeError())

        sched.startService()
        self.assertEqual(1, len(self.flushLoggedErrors(RuntimeError)))
        self.assertFalse(sched.isActive())

    def test_activation_activate_fails(self):
        sched = self.makeScheduler(name='n', builderNames=['a'])
        sched.clock = task.Clock()

        def activate():
            raise RuntimeError('oh noes')
        sched.activate = activate

        sched.startService()
        self.assertEqual(1, len(self.flushLoggedErrors(RuntimeError)))

    @defer.inlineCallbacks
    def do_addBuildsetForSourceStampsWithDefaults(self, codebases,
                                                  sourcestamps,
                                                  exp_sourcestamps):
        sched = self.makeScheduler(name='n', builderNames=['b'],
                                   codebases=codebases)
        bsid, brids = yield sched.addBuildsetForSourceStampsWithDefaults(
            reason='power', sourcestamps=sourcestamps, waited_for=False)
        self.assertEqual((bsid, brids), self.exp_bsid_brids)
        call = self.master.data.updates.addBuildset.mock_calls[0]

        def sourceStampKey(sourceStamp):
            repository = sourceStamp.get('repository', '')
            if repository is None:
                repository = ''
            branch = sourceStamp.get('branch', '') if not None else ''
            if branch is None:
                branch = ''
            return (repository, branch)

        self.assertEqual(sorted(call[2]['sourcestamps'], key=sourceStampKey),
                         sorted(exp_sourcestamps, key=sourceStampKey))

    def test_addBuildsetForSourceStampsWithDefaults(self):
        codebases = {
            'cbA': dict(repository='svn://A..', branch='stable',
                        revision='13579'),
            'cbB': dict(repository='svn://B..', branch='stable',
                        revision='24680')
        }
        sourcestamps = [
            {'codebase': 'cbA', 'branch': 'AA'},
            {'codebase': 'cbB', 'revision': 'BB'},
        ]
        exp_sourcestamps = [
            {'repository': 'svn://B..', 'branch': 'stable',
             'revision': 'BB', 'codebase': 'cbB', 'project': ''},
            {'repository': 'svn://A..', 'branch': 'AA', 'project': '',
             'revision': '13579', 'codebase': 'cbA'},
        ]
        return self.do_addBuildsetForSourceStampsWithDefaults(
            codebases, sourcestamps, exp_sourcestamps)

    def test_addBuildsetForSourceStampsWithDefaults_fill_in_codebases(self):
        codebases = {
            'cbA': dict(repository='svn://A..', branch='stable',
                        revision='13579'),
            'cbB': dict(repository='svn://B..', branch='stable',
                        revision='24680')
        }
        sourcestamps = [
            {'codebase': 'cbA', 'branch': 'AA'},
        ]
        exp_sourcestamps = [
            {'repository': 'svn://B..', 'branch': 'stable',
             'revision': '24680', 'codebase': 'cbB', 'project': ''},
            {'repository': 'svn://A..', 'branch': 'AA', 'project': '',
             'revision': '13579', 'codebase': 'cbA'},
        ]
        return self.do_addBuildsetForSourceStampsWithDefaults(
            codebases, sourcestamps, exp_sourcestamps)

    def test_addBuildsetForSourceStampsWithDefaults_no_repository(self):
        exp_sourcestamps = [
            {'repository': '', 'branch': None,
             'revision': None, 'codebase': '', 'project': ''},
        ]
        return self.do_addBuildsetForSourceStampsWithDefaults(
            {'': {}}, [], exp_sourcestamps)

    def test_addBuildsetForSourceStamps_unknown_codbases(self):
        codebases = {}
        sourcestamps = [
            {'codebase': 'cbA', 'branch': 'AA'},
            {'codebase': 'cbB', 'revision': 'BB'},
        ]
        exp_sourcestamps = [
            {'branch': None, 'revision': 'BB', 'codebase': 'cbB',
             'project': '', 'repository': ''},
            {'branch': 'AA', 'revision': None, 'codebase': 'cbA',
             'project': '', 'repository': ''},
        ]
        return self.do_addBuildsetForSourceStampsWithDefaults(
            codebases, sourcestamps, exp_sourcestamps)

    @defer.inlineCallbacks
    def test_addBuildsetForChanges_one_change(self):
        sched = self.makeScheduler(name='n', builderNames=['b'])
        self.db.insertTestData([
            fakedb.Change(changeid=13, sourcestampid=234),
        ])
        bsid, brids = yield sched.addBuildsetForChanges(reason='power',
                                                        waited_for=False, changeids=[13])
        self.assertEqual((bsid, brids), self.exp_bsid_brids)
        self.master.data.updates.addBuildset.assert_called_with(
            waited_for=False,
            builderids=[1],
            external_idstring=None,
            properties={
                'scheduler': ('n', 'Scheduler'),
            },
            reason='power',
            scheduler='n',
            sourcestamps=[234])

    @defer.inlineCallbacks
    def test_addBuildsetForChanges_properties(self):
        sched = self.makeScheduler(name='n', builderNames=['c'])
        self.db.insertTestData([
            fakedb.Change(changeid=14, sourcestampid=234),
        ])
        bsid, brids = yield sched.addBuildsetForChanges(reason='downstream',
                                                        waited_for=False, changeids=[14])
        self.assertEqual((bsid, brids), self.exp_bsid_brids)
        self.master.data.updates.addBuildset.assert_called_with(
            waited_for=False,
            builderids=[1],
            external_idstring=None,
            properties={
                'scheduler': ('n', 'Scheduler'),
            },
            reason='downstream',
            scheduler='n',
            sourcestamps=[234])

    @defer.inlineCallbacks
    def test_addBuildsetForChanges_properties_with_virtual_builders(self):
        sched = self.makeScheduler(name='n', builderNames=['c'], properties={
            'virtual_builder_name': Interpolate("myproject-%(src::branch)s")
        })
        self.db.insertTestData([
            fakedb.SourceStamp(id=234, branch='dev1', project="linux"),
            fakedb.Change(changeid=14, sourcestampid=234, branch="dev1"),
        ])
        bsid, brids = yield sched.addBuildsetForChanges(reason='downstream',
                                                        waited_for=False, changeids=[14])
        self.assertEqual((bsid, brids), self.exp_bsid_brids)
        self.master.data.updates.addBuildset.assert_called_with(
            waited_for=False,
            builderids=[1],
            external_idstring=None,
            properties={
                'virtual_builder_name': ("myproject-dev1", "Scheduler"),
                'scheduler': ('n', 'Scheduler'),
            },
            reason='downstream',
            scheduler='n',
            sourcestamps=[234])

    @defer.inlineCallbacks
    def test_addBuildsetForChanges_multiple_changes_same_codebase(self):
        # This is a test for backwards compatibility
        # Changes from different repositories come together in one build
        sched = self.makeScheduler(name='n', builderNames=['b', 'c'],
                                   codebases={'cb': {'repository': 'http://repo'}})
        # No codebaseGenerator means all changes have codebase == ''
        self.db.insertTestData([
            fakedb.Change(changeid=13, codebase='cb', sourcestampid=12),
            fakedb.Change(changeid=14, codebase='cb', sourcestampid=11),
            fakedb.Change(changeid=15, codebase='cb', sourcestampid=10),
        ])

        # note that the changeids are given out of order here; it should still
        # use the most recent
        bsid, brids = yield sched.addBuildsetForChanges(reason='power',
                                                        waited_for=False, changeids=[14, 15, 13])
        self.assertEqual((bsid, brids), self.exp_bsid_brids)
        self.master.data.updates.addBuildset.assert_called_with(
            waited_for=False,
            builderids=[1, 2],
            external_idstring=None,
            properties={
                'scheduler': ('n', 'Scheduler'),
            },
            reason='power',
            scheduler='n',
            sourcestamps=[10])  # sourcestampid from greatest changeid

    @defer.inlineCallbacks
    def test_addBuildsetForChanges_codebases_set_multiple_codebases(self):
        codebases = {'cbA': dict(repository='svn://A..',
                                 branch='stable',
                                 revision='13579'),
                     'cbB': dict(
                         repository='svn://B..',
                         branch='stable',
                         revision='24680'),
                     'cbC': dict(
                         repository='svn://C..',
                         branch='stable',
                         revision='12345'),
                     'cbD': dict(
                         repository='svn://D..')}
        # Scheduler gets codebases that can be used to create extra sourcestamps
        # for repositories that have no changes
        sched = self.makeScheduler(name='n', builderNames=['b', 'c'],
                                   codebases=codebases)
        self.db.insertTestData([
            fakedb.Change(changeid=12, codebase='cbA', sourcestampid=912),
            fakedb.Change(changeid=13, codebase='cbA', sourcestampid=913),
            fakedb.Change(changeid=14, codebase='cbA', sourcestampid=914),
            fakedb.Change(changeid=15, codebase='cbB', sourcestampid=915),
            fakedb.Change(changeid=16, codebase='cbB', sourcestampid=916),
            fakedb.Change(changeid=17, codebase='cbB', sourcestampid=917),
            # note: no changes for cbC or cbD
        ])

        # note that the changeids are given out of order here; it should still
        # use the most recent for each codebase
        bsid, brids = yield sched.addBuildsetForChanges(reason='power', waited_for=True,
                                                        changeids=[14, 12, 17, 16, 13, 15])
        self.assertEqual((bsid, brids), self.exp_bsid_brids)

        self.master.data.updates.addBuildset.assert_called_with(
            waited_for=True,
            builderids=[1, 2],
            external_idstring=None,
            reason='power',
            scheduler='n',
            properties={
                'scheduler': ('n', 'Scheduler'),
            },
            sourcestamps=[914,
                          917,
                          dict(branch='stable', repository='svn://C..',
                               codebase='cbC', project='', revision='12345'),
                          dict(branch=None, repository='svn://D..', codebase='cbD',
                               project='', revision=None)
                          ]
        )

    @defer.inlineCallbacks
    def test_addBuildsetForSourceStamp(self):
        sched = self.makeScheduler(name='n', builderNames=['b'])
        sourcestamps = [91, {'sourcestamp': True}]
        bsid, brids = yield sched.addBuildsetForSourceStamps(reason='whynot', waited_for=False,
                                                             sourcestamps=sourcestamps)
        self.assertEqual((bsid, brids), self.exp_bsid_brids)
        self.master.data.updates.addBuildset.assert_called_with(
            waited_for=False,
            builderids=[1],
            external_idstring=None,
            reason='whynot',
            scheduler='n',
            properties={
                'scheduler': ('n', 'Scheduler'),
            },
            sourcestamps=[91, {'sourcestamp': True}])

    @defer.inlineCallbacks
    def test_addBuildsetForSourceStamp_explicit_builderNames(self):
        sched = self.makeScheduler(name='n', builderNames=['b', 'x', 'y'])
        bsid, brids = yield sched.addBuildsetForSourceStamps(reason='whynot',
                                                             waited_for=True,
                                                             sourcestamps=[
                                                                 91, {'sourcestamp': True}],
                                                             builderNames=['x', 'y'])
        self.assertEqual((bsid, brids), self.exp_bsid_brids)
        self.master.data.updates.addBuildset.assert_called_with(
            waited_for=True,
            builderids=[2, 3],
            external_idstring=None,
            reason='whynot',
            scheduler='n',
            properties={
                'scheduler': ('n', 'Scheduler'),
            },
            sourcestamps=[91, {'sourcestamp': True}])

    @defer.inlineCallbacks
    def test_addBuildsetForSourceStamp_properties(self):
        props = properties.Properties(xxx="yyy")
        sched = self.makeScheduler(name='n', builderNames=['b'])
        bsid, brids = yield sched.addBuildsetForSourceStamps(reason='whynot',
                                                             waited_for=False,
                                                             sourcestamps=[91], properties=props)
        self.assertEqual((bsid, brids), self.exp_bsid_brids)
        self.master.data.updates.addBuildset.assert_called_with(
            waited_for=False,
            builderids=[1],
            external_idstring=None,
            properties={
                'xxx': ('yyy', 'TEST'),
                'scheduler': ('n', 'Scheduler')},
            reason='whynot',
            scheduler='n',
            sourcestamps=[91])

    @defer.inlineCallbacks
    def test_addBuildsetForSourceStamp_combine_change_properties(self):
        sched = self.makeScheduler()

        self.master.db.insertTestData([
            fakedb.SourceStamp(id=98, branch='stable'),
            fakedb.Change(changeid=25, sourcestampid=98, branch='stable'),
            fakedb.ChangeProperty(changeid=25, property_name='color',
                                  property_value='["pink","Change"]'),
        ])

        bsid, brids = yield sched.addBuildsetForSourceStamps(reason='whynot',
                                                             waited_for=False,
                                                             sourcestamps=[98])
        self.assertEqual((bsid, brids), self.exp_bsid_brids)
        self.master.data.updates.addBuildset.assert_called_with(
            waited_for=False,
            builderids=[1, 2],
            external_idstring=None,
            properties={
                'scheduler': ('testsched', 'Scheduler'),
                'color': ('pink', 'Change')},
            reason='whynot',
            scheduler='testsched',
            sourcestamps=[98])

    @defer.inlineCallbacks
    def test_addBuildsetForSourceStamp_renderable_builderNames(self):
        @properties.renderer
        def names(props):
            if props.changes[0]['branch'] == 'stable':
                return ['c']
            elif props.changes[0]['branch'] == 'unstable':
                return ['a', 'b']
            return None

        sched = self.makeScheduler(name='n', builderNames=names)

        self.master.db.insertTestData([
            fakedb.Builder(id=1, name='a'),
            fakedb.Builder(id=2, name='b'),
            fakedb.Builder(id=3, name='c'),
            fakedb.SourceStamp(id=98, branch='stable'),
            fakedb.SourceStamp(id=99, branch='unstable'),
            fakedb.Change(changeid=25, sourcestampid=98, branch='stable'),
            fakedb.Change(changeid=26, sourcestampid=99, branch='unstable'),
        ])

        bsid, brids = yield sched.addBuildsetForSourceStamps(reason='whynot',
                                                             waited_for=False,
                                                             sourcestamps=[98])
        self.assertEqual((bsid, brids), self.exp_bsid_brids)
        self.master.data.updates.addBuildset.assert_called_with(
            waited_for=False,
            builderids=[3],
            external_idstring=None,
            properties={
                'scheduler': ('n', 'Scheduler')},
            reason='whynot',
            scheduler='n',
            sourcestamps=[98])

        bsid, brids = yield sched.addBuildsetForSourceStamps(reason='because',
                                                             waited_for=False,
                                                             sourcestamps=[99])
        self.assertEqual((bsid, brids), self.exp_bsid_brids)
        self.master.data.updates.addBuildset.assert_called_with(
            waited_for=False,
            builderids=[1, 2],
            external_idstring=None,
            properties={
                'scheduler': ('n', 'Scheduler')},
            reason='because',
            scheduler='n',
            sourcestamps=[99])

    @defer.inlineCallbacks
    def test_addBuildsetForSourceStamp_list_of_renderable_builderNames(self):
        names = ['a', 'b', properties.Interpolate('%(prop:extra_builder)s')]
        sched = self.makeScheduler(name='n', builderNames=names)

        self.master.db.insertTestData([
            fakedb.Builder(id=1, name='a'),
            fakedb.Builder(id=2, name='b'),
            fakedb.Builder(id=3, name='c'),
            fakedb.SourceStamp(id=98, branch='stable'),
            fakedb.Change(changeid=25, sourcestampid=98, branch='stable'),
            fakedb.ChangeProperty(changeid=25, property_name='extra_builder',
                                  property_value='["c","Change"]'),
        ])

        bsid, brids = yield sched.addBuildsetForSourceStamps(reason='whynot',
                                                             waited_for=False,
                                                             sourcestamps=[98])
        self.assertEqual((bsid, brids), self.exp_bsid_brids)
        self.master.data.updates.addBuildset.assert_called_with(
            waited_for=False,
            builderids=[1, 2, 3],
            external_idstring=None,
            properties={
                'scheduler': ('n', 'Scheduler'),
                'extra_builder': ('c', 'Change')},
            reason='whynot',
            scheduler='n',
            sourcestamps=[98])

    def test_signature_addBuildsetForChanges(self):
        sched = self.makeScheduler(builderNames=['xxx'])

        @self.assertArgSpecMatches(
            sched.addBuildsetForChanges,  # Real
            self.fake_addBuildsetForChanges,  # Real
        )
        def addBuildsetForChanges(self, waited_for=False, reason='',
                                  external_idstring=None, changeids=None, builderNames=None,
                                  properties=None,
                                  **kw):
            pass

    def test_signature_addBuildsetForSourceStamps(self):
        sched = self.makeScheduler(builderNames=['xxx'])

        @self.assertArgSpecMatches(
            sched.addBuildsetForSourceStamps,  # Real
            self.fake_addBuildsetForSourceStamps,  # Fake
        )
        def addBuildsetForSourceStamps(self, waited_for=False, sourcestamps=None,
                                       reason='', external_idstring=None, properties=None,
                                       builderNames=None, **kw):
            pass

    def test_signature_addBuildsetForSourceStampsWithDefaults(self):
        sched = self.makeScheduler(builderNames=['xxx'])

        @self.assertArgSpecMatches(
            sched.addBuildsetForSourceStampsWithDefaults,  # Real
            self.fake_addBuildsetForSourceStampsWithDefaults,  # Fake
        )
        def addBuildsetForSourceStampsWithDefaults(self, reason, sourcestamps=None,
                                                   waited_for=False, properties=None,
                                                   builderNames=None, **kw):
            pass



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.internet import defer
from twisted.trial import reporter
from twisted.trial import unittest

from buildbot.test.runprocess import ExpectMasterShell
from buildbot.test.runprocess import MasterRunProcessMixin
from buildbot.util import runprocess


class TestRunprocessMixin(unittest.TestCase):

    def run_test_method(self, method):
        class TestCase(MasterRunProcessMixin, unittest.TestCase):

            def setUp(self):
                self.setup_master_run_process()

            def runTest(self):
                return method(self)

        self.testcase = TestCase()
        result = reporter.TestResult()
        self.testcase.run(result)  # This blocks
        return result

    def assert_test_failure(self, result, expected_failure):
        self.assertEqual(result.errors, [])
        self.assertEqual(len(result.failures), 1)
        self.assertTrue(result.failures[0][1].check(unittest.FailTest))
        if expected_failure:
            self.assertSubstring(expected_failure, result.failures[0][1].getErrorMessage())

    def assert_successful(self, result):
        if not result.wasSuccessful():
            output = 'expected success'
            if result.failures:
                output += (f'\ntest failed: {result.failures[0][1].getErrorMessage()}')
            if result.errors:
                output += (f'\nerrors: {[error[1].value for error in result.errors]}')
            raise self.failureException(output)

        self.assertTrue(result.wasSuccessful())

    def test_patch(self):
        original_run_process = runprocess.run_process

        def method(testcase):
            testcase.expect_commands()
            self.assertEqual(runprocess.run_process, testcase.patched_run_process)

        result = self.run_test_method(method)
        self.assert_successful(result)
        self.assertEqual(runprocess.run_process, original_run_process)

    def test_method_chaining(self):
        expect = ExpectMasterShell('command')
        self.assertEqual(expect, expect.exit(0))
        self.assertEqual(expect, expect.stdout(b"output"))
        self.assertEqual(expect, expect.stderr(b"error"))

    def test_run_process_one_command_only_rc(self):

        @defer.inlineCallbacks
        def method(testcase):
            testcase.expect_commands(
                ExpectMasterShell(["command"]).stdout(b'stdout').stderr(b'stderr'))
            res = yield runprocess.run_process(None, ["command"],
                                               collect_stdout=False, collect_stderr=False)
            self.assertEqual(res, 0)
            testcase.assert_all_commands_ran()

        result = self.run_test_method(method)
        self.assert_successful(result)

    def test_run_process_one_command_only_rc_stdout(self):

        @defer.inlineCallbacks
        def method(testcase):
            testcase.expect_commands(
                ExpectMasterShell(["command"]).stdout(b'stdout').stderr(b'stderr'))
            res = yield runprocess.run_process(None, ["command"],
                                               collect_stdout=True, collect_stderr=False)
            self.assertEqual(res, (0, b'stdout'))
            testcase.assert_all_commands_ran()

        result = self.run_test_method(method)
        self.assert_successful(result)

    def test_run_process_one_command_with_rc_stderr(self):

        @defer.inlineCallbacks
        def method(testcase):
            testcase.expect_commands(
                ExpectMasterShell(["command"]).stdout(b'stdout').stderr(b'stderr'))
            res = yield runprocess.run_process(None, ["command"],
                                               collect_stdout=False, collect_stderr=True)
            self.assertEqual(res, (0, b'stderr'))
            testcase.assert_all_commands_ran()

        result = self.run_test_method(method)
        self.assert_successful(result)

    def test_run_process_one_command_with_rc_stdout_stderr(self):

        @defer.inlineCallbacks
        def method(testcase):
            testcase.expect_commands(
                ExpectMasterShell(["command"]).stdout(b'stdout').stderr(b'stderr'))
            res = yield runprocess.run_process(None, ["command"])
            self.assertEqual(res, (0, b'stdout', b'stderr'))
            testcase.assert_all_commands_ran()

        result = self.run_test_method(method)
        self.assert_successful(result)

    def test_run_process_expect_two_run_one(self):

        @defer.inlineCallbacks
        def method(testcase):
            testcase.expect_commands(ExpectMasterShell(["command"]))
            testcase.expect_commands(ExpectMasterShell(["command2"]))
            res = yield runprocess.run_process(None, ["command"])
            self.assertEqual(res, (0, b'', b''))
            testcase.assert_all_commands_ran()

        result = self.run_test_method(method)
        self.assert_test_failure(result, "assert all expected commands were run")

    def test_run_process_wrong_command(self):

        @defer.inlineCallbacks
        def method(testcase):
            testcase.expect_commands(ExpectMasterShell(["command2"]))
            yield runprocess.run_process(None, ["command"])

        result = self.run_test_method(method)
        self.assert_test_failure(result, "unexpected command run")
        # assert we have a meaningful message
        self.assert_test_failure(result, "command2")

    def test_run_process_wrong_args(self):

        @defer.inlineCallbacks
        def method(testcase):
            testcase.expect_commands(ExpectMasterShell(["command", "arg"]))
            yield runprocess.run_process(None, ["command", "otherarg"])
            testcase.assert_all_commands_ran()

        result = self.run_test_method(method)
        self.assert_test_failure(result, "unexpected command run")

    def test_run_process_missing_path(self):

        @defer.inlineCallbacks
        def method(testcase):
            testcase.expect_commands(ExpectMasterShell(["command"]).workdir("/home"))
            yield runprocess.run_process(None, ["command"])
            testcase.assert_all_commands_ran()

        result = self.run_test_method(method)
        self.assert_test_failure(result, "unexpected command run")

    def test_run_process_wrong_path(self):

        @defer.inlineCallbacks
        def method(testcase):
            testcase.expect_commands(ExpectMasterShell(["command", "arg"]).workdir("/home"))
            yield runprocess.run_process(None, ["command"], workdir="/path")
            testcase.assert_all_commands_ran()

        result = self.run_test_method(method)
        self.assert_test_failure(result, "unexpected command run")

    def test_run_process_not_current_path(self):

        @defer.inlineCallbacks
        def method(testcase):
            testcase.expect_commands(ExpectMasterShell(["command", "arg"]))
            yield runprocess.run_process(None, ["command"], workdir="/path")
            testcase.assert_all_commands_ran()

        result = self.run_test_method(method)
        self.assert_test_failure(result, "unexpected command run")

    def test_run_process_error_output(self):

        @defer.inlineCallbacks
        def method(testcase):
            testcase.expect_commands(ExpectMasterShell(["command"]).stderr(b"some test"))
            res = yield runprocess.run_process(None, ["command"], collect_stderr=False,
                                               stderr_is_error=True)
            self.assertEqual(res, (-1, b''))
            testcase.assert_all_commands_ran()

        result = self.run_test_method(method)
        self.assert_successful(result)

    def test_run_process_nonzero_exit(self):

        @defer.inlineCallbacks
        def method(testcase):
            testcase.expect_commands(ExpectMasterShell(["command"]).exit(1))
            res = yield runprocess.run_process(None, ["command"])
            self.assertEqual(res, (1, b'', b''))
            testcase.assert_all_commands_ran()

        result = self.run_test_method(method)
        self.assert_successful(result)

    def test_run_process_environ_success(self):

        @defer.inlineCallbacks
        def method(testcase):
            testcase.expect_commands(ExpectMasterShell(["command"]))
            testcase.add_run_process_expect_env({'key': 'value'})
            res = yield runprocess.run_process(None, ["command"], env={'key': 'value'})
            self.assertEqual(res, (0, b'', b''))
            testcase.assert_all_commands_ran()

        result = self.run_test_method(method)
        self.assert_successful(result)

    def test_run_process_environ_wrong_value(self):

        @defer.inlineCallbacks
        def method(testcase):
            testcase.expect_commands(ExpectMasterShell(["command"]))
            testcase.add_run_process_expect_env({'key': 'value'})
            yield runprocess.run_process(None, ["command"], env={'key': 'wrongvalue'})
            testcase.assert_all_commands_ran()

        result = self.run_test_method(method)
        self.assert_test_failure(result, "Expected environment to have key = 'value'")

    def test_run_process_environ_missing(self):
        @defer.inlineCallbacks
        def method(testcase):
            testcase.expect_commands(ExpectMasterShell(["command"]))
            testcase.add_run_process_expect_env({'key': 'value'})
            d = runprocess.run_process(None, ["command"])
            return d
        result = self.run_test_method(method)
        self.assert_test_failure(result, "Expected environment to have key = 'value'")



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import os
import textwrap

import sqlalchemy as sa

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.scripts import cleanupdb
from buildbot.test.fake import fakemaster
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.unit.db import test_logs
from buildbot.test.util import db
from buildbot.test.util import dirs
from buildbot.test.util import misc

try:
    import lz4
    [lz4]
    hasLz4 = True
except ImportError:
    hasLz4 = False


def mkconfig(**kwargs):
    config = dict(quiet=False, basedir=os.path.abspath('basedir'), force=True)
    config.update(kwargs)
    return config


def patch_environ(case, key, value):
    """
    Add an environment variable for the duration of a test.
    """
    old_environ = os.environ.copy()

    def cleanup():
        os.environ.clear()
        os.environ.update(old_environ)
    os.environ[key] = value
    case.addCleanup(cleanup)


class TestCleanupDb(misc.StdoutAssertionsMixin, dirs.DirsMixin,
                    TestReactorMixin, unittest.TestCase):

    def setUp(self):
        self.setup_test_reactor()
        self.origcwd = os.getcwd()
        self.setUpDirs('basedir')
        with open(os.path.join('basedir', 'buildbot.tac'), 'wt', encoding='utf-8') as f:
            f.write(textwrap.dedent("""
                from twisted.application import service
                application = service.Application('buildmaster')
            """))
        self.setUpStdoutAssertions()
        self.ensureNoSqliteMemory()

    def tearDown(self):
        os.chdir(self.origcwd)
        self.tearDownDirs()

    def ensureNoSqliteMemory(self):
        # test may use mysql or pg if configured in env
        envkey = "BUILDBOT_TEST_DB_URL"
        if envkey not in os.environ or os.environ[envkey] == 'sqlite://':

            patch_environ(self, envkey, "sqlite:///" + os.path.join(
                self.origcwd, "basedir", "state.sqlite"))

    def createMasterCfg(self, extraconfig=""):
        os.chdir(self.origcwd)
        with open(os.path.join('basedir', 'master.cfg'), 'wt', encoding='utf-8') as f:
            f.write(textwrap.dedent(f"""
                from buildbot.plugins import *
                c = BuildmasterConfig = dict()
                c['db_url'] = {repr(os.environ["BUILDBOT_TEST_DB_URL"])}
                c['buildbotNetUsageData'] = None
                c['multiMaster'] = True  # don't complain for no builders
                {extraconfig}
            """))

    @defer.inlineCallbacks
    def test_cleanup_not_basedir(self):
        res = yield cleanupdb._cleanupDatabase(mkconfig(basedir='doesntexist'))
        self.assertEqual(res, 1)
        self.assertInStdout('invalid buildmaster directory')

    @defer.inlineCallbacks
    def test_cleanup_bad_config(self):
        res = yield cleanupdb._cleanupDatabase(mkconfig(basedir='basedir'))
        self.assertEqual(res, 1)
        self.assertInStdout("master.cfg' does not exist")

    @defer.inlineCallbacks
    def test_cleanup_bad_config2(self):

        self.createMasterCfg(extraconfig="++++ # syntaxerror")
        res = yield cleanupdb._cleanupDatabase(mkconfig(basedir='basedir'))
        self.assertEqual(res, 1)
        self.assertInStdout(
            "encountered a SyntaxError while parsing config file:")
        # config logs an error via log.err, we must eat it or trial will
        # complain
        self.flushLoggedErrors()

    def assertDictAlmostEqual(self, d1, d2):
        # The test shows each methods return different size
        # but we still make a fuzzy comparison to resist if underlying libraries
        # improve efficiency
        self.assertEqual(len(d1), len(d2))
        for k in d2.keys():
            self.assertApproximates(d1[k], d2[k], 10)


class TestCleanupDbRealDb(db.RealDatabaseWithConnectorMixin, TestCleanupDb):

    @defer.inlineCallbacks
    def setUp(self):
        yield super().setUp()

        table_names = [
            'logs', 'logchunks', 'steps', 'builds', 'builders',
            'masters', 'buildrequests', 'buildsets', 'workers'
        ]

        self.master = fakemaster.make_master(self, wantRealReactor=True)
        yield self.setUpRealDatabaseWithConnector(self.master, table_names=table_names)

    @defer.inlineCallbacks
    def tearDown(self):
        yield self.tearDownRealDatabaseWithConnector()

    @defer.inlineCallbacks
    def test_cleanup(self):
        # we reuse the fake db background data from db.logs unit tests
        yield self.insertTestData(test_logs.Tests.backgroundData)

        # insert a log with lots of redundancy
        LOGDATA = "xx\n" * 2000
        logid = yield self.master.db.logs.addLog(102, "x", "x", "s")
        yield self.master.db.logs.appendLog(logid, LOGDATA)

        # test all methods
        lengths = {}
        for mode in self.master.db.logs.COMPRESSION_MODE:
            if mode == "lz4" and not hasLz4:
                # ok.. lz4 is not installed, don't fail
                lengths["lz4"] = 40
                continue
            # create a master.cfg with different compression method
            self.createMasterCfg(f"c['logCompressionMethod'] = '{mode}'")
            res = yield cleanupdb._cleanupDatabase(mkconfig(basedir='basedir'))
            self.assertEqual(res, 0)

            # make sure the compression don't change the data we can retrieve
            # via api
            res = yield self.master.db.logs.getLogLines(logid, 0, 2000)
            self.assertEqual(res, LOGDATA)

            # retrieve the actual data size in db using raw sqlalchemy
            def thd(conn):
                tbl = self.master.db.model.logchunks
                q = sa.select([sa.func.sum(sa.func.length(tbl.c.content))])
                q = q.where(tbl.c.logid == logid)
                return conn.execute(q).fetchone()[0]
            lengths[mode] = yield self.master.db.pool.do(thd)

        self.assertDictAlmostEqual(
            lengths, {'raw': 5999, 'bz2': 44, 'lz4': 40, 'gz': 31})



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members
#
# N.B.: don't import anything that might pull in a reactor yet. Some of our
# subcommands want to load modules that need the gtk reactor.
#
# Also don't forget to mirror your changes on command-line options in manual
# pages and reStructuredText documentation.

import getpass
import sys
import textwrap

import sqlalchemy as sa

from twisted.python import reflect
from twisted.python import usage

import buildbot
from buildbot.scripts import base
from buildbot.util import check_functional_environment

# Note that the terms 'options' and 'config' are used interchangeably here - in
# fact, they are interchanged several times.  Caveat legator.


def validateMasterOption(master):
    """
    Validate master (-m, --master) command line option.

    Checks that option is a string of the 'hostname:port' form, otherwise
    raises an UsageError exception.

    @type  master: string
    @param master: master option

    @raise usage.UsageError: on invalid master option
    """
    try:
        _, port = master.split(":")
        port = int(port)
    except (TypeError, ValueError) as e:
        raise usage.UsageError("master must have the form 'hostname:port'") from e


class UpgradeMasterOptions(base.BasedirMixin, base.SubcommandOptions):
    subcommandFunction = "buildbot.scripts.upgrade_master.upgradeMaster"
    optFlags = [
        ["quiet", "q", "Do not emit the commands being run"],
        ["develop", "d", "link to buildbot dir rather than copy, with no "
                         "JS optimization (UNIX only)"],
        ["replace", "r", "Replace any modified files without confirmation."],
    ]
    optParameters = [
    ]

    def getSynopsis(self):
        return "Usage:    buildbot upgrade-master [options] [<basedir>]"

    longdesc = textwrap.dedent("""
    This command takes an existing buildmaster working directory and
    adds/modifies the files there to work with the current version of
    buildbot. When this command is finished, the buildmaster directory should
    look much like a brand-new one created by the 'create-master' command.

    Use this after you've upgraded your buildbot installation and before you
    restart the buildmaster to use the new version.

    If you have modified the files in your working directory, this command
    will leave them untouched, but will put the new recommended contents in a
    .new file (for example, if index.html has been modified, this command
    will create index.html.new). You can then look at the new version and
    decide how to merge its contents into your modified file.

    When upgrading the database, this command uses the database specified in
    the master configuration file.  If you wish to use a database other than
    the default (sqlite), be sure to set that parameter before upgrading.
    """)


class CreateMasterOptions(base.BasedirMixin, base.SubcommandOptions):
    subcommandFunction = "buildbot.scripts.create_master.createMaster"
    optFlags = [
        ["quiet", "q", "Do not emit the commands being run"],
        ["force", "f",
         "Re-use an existing directory (will not overwrite master.cfg file)"],
        ["relocatable", "r",
         "Create a relocatable buildbot.tac"],
        ["develop", "d", "link to buildbot dir rather than copy, with no "
                         "JS optimization (UNIX only)"],
        ["no-logrotate", "n",
         "Do not permit buildmaster rotate logs by itself"]
    ]
    optParameters = [
        ["config", "c", "master.cfg", "name of the buildmaster config file"],
        ["log-size", "s", 10000000,
         "size at which to rotate twisted log files", int],
        ["log-count", "l", 10,
         "limit the number of kept old twisted log files"],
        ["db", None, "sqlite:///state.sqlite",
         "which DB to use for scheduler/status state. See below for syntax."],
    ]

    def getSynopsis(self):
        return "Usage:    buildbot create-master [options] [<basedir>]"

    longdesc = textwrap.dedent("""
    This command creates a buildmaster working directory and buildbot.tac file.
    The master will live in <basedir> (defaults to the current directory)
    and create various files there.
    If --relocatable is given, then the resulting buildbot.tac file will be
    written such that its containing directory is assumed to be the basedir.
    This is generally a good idea.

    At runtime, the master will read a configuration file (named
    'master.cfg' by default) in its basedir. This file should contain python
    code which eventually defines a dictionary named 'BuildmasterConfig'.
    The elements of this dictionary are used to configure the Buildmaster.
    See doc/config.xhtml for details about what can be controlled through
    this interface.

    The --db string is evaluated to build the DB object, which specifies
    which database the buildmaster should use to hold scheduler state and
    status information. The default (which creates an SQLite database in
    BASEDIR/state.sqlite) is equivalent to:

      --db='sqlite:///state.sqlite'

    To use a remote MySQL database instead, use something like:

      --db='mysql://bbuser:bbpasswd@dbhost/bbdb'
    The --db string is stored verbatim in the master.cfg.sample file, and
    evaluated at 'buildbot start' time to pass a DBConnector instance into
    the newly-created BuildMaster object.
    """)

    def postOptions(self):
        super().postOptions()

        # validate 'log-count' parameter
        if self['log-count'] == 'None':
            self['log-count'] = None
        else:
            try:
                self['log-count'] = int(self['log-count'])
            except ValueError as e:
                raise usage.UsageError(
                    "log-count parameter needs to be an int or None") from e

        # validate 'db' parameter
        try:
            # check if sqlalchemy will be able to parse specified URL
            sa.engine.url.make_url(self['db'])
        except sa.exc.ArgumentError as e:
            raise usage.UsageError(f"could not parse database URL '{self['db']}'") from e


class StopOptions(base.BasedirMixin, base.SubcommandOptions):
    subcommandFunction = "buildbot.scripts.stop.stop"
    optFlags = [
        ["quiet", "q", "Do not emit the commands being run"],
        ["clean", "c", "Clean shutdown master"],
        ["no-wait", None, "Don't wait for complete master shutdown"],
    ]

    def getSynopsis(self):
        return "Usage:    buildbot stop [<basedir>]"


class RestartOptions(base.BasedirMixin, base.SubcommandOptions):
    subcommandFunction = "buildbot.scripts.restart.restart"
    optFlags = [
        ['quiet', 'q', "Don't display startup log messages"],
        ['nodaemon', None, "Don't daemonize (stay in foreground)"],
        ["clean", "c", "Clean shutdown master"],
    ]
    optParameters = [
        ['start_timeout', None, None,
         'The amount of time the script waits for the master to restart until '
         'it declares the operation as failure'],
    ]

    def getSynopsis(self):
        return "Usage:    buildbot restart [<basedir>]"


class StartOptions(base.BasedirMixin, base.SubcommandOptions):
    subcommandFunction = "buildbot.scripts.start.start"
    optFlags = [
        ['quiet', 'q', "Don't display startup log messages"],
        ['nodaemon', None, "Don't daemonize (stay in foreground)"],
    ]
    optParameters = [
        ['start_timeout', None, None,
         'The amount of time the script waits for the master to start until it '
         'declares the operation as failure'],
    ]

    def getSynopsis(self):
        return "Usage:    buildbot start [<basedir>]"


class ReconfigOptions(base.BasedirMixin, base.SubcommandOptions):
    subcommandFunction = "buildbot.scripts.reconfig.reconfig"
    optFlags = [
        ['quiet', 'q', "Don't display log messages about reconfiguration"],
    ]

    optParameters = [
        ['progress_timeout', None, None,
         'The amount of time the script waits for messages in the logs that indicate progress.'],
    ]

    def getSynopsis(self):
        return "Usage:    buildbot reconfig [<basedir>]"


class SendChangeOptions(base.SubcommandOptions):
    subcommandFunction = "buildbot.scripts.sendchange.sendchange"

    def __init__(self):
        super().__init__()
        self['properties'] = {}

    optParameters = [
        ("master", "m", None,
         "Location of the buildmaster's PBChangeSource (host:port)"),
        # deprecated in 0.8.3; remove in 0.8.5 (bug #1711)
        ("auth", "a", 'change:changepw',
         "Authentication token - username:password, or prompt for password"),
        ("who", "W", None, "Author of the commit"),
        ("repository", "R", '', "Repository specifier"),
        ("vc", "s", None, "The VC system in use, one of: cvs, svn, darcs, hg, "
         "bzr, git, mtn, p4"),
        ("project", "P", '', "Project specifier"),
        ("branch", "b", None, "Branch specifier"),
        ("category", "C", None, "Category of repository"),
        ("codebase", None, None,
            "Codebase this change is in (requires 0.8.7 master or later)"),
        ("revision", "r", None, "Revision specifier"),
        ("revision_file", None, None, "Filename containing revision spec"),
        ("property", "p", None,
         "A property for the change, in the format: name:value"),
        ("comments", "c", None, "log message"),
        ("logfile", "F", None,
         "Read the log messages from this file (- for stdin)"),
        ("when", "w", None, "timestamp to use as the change time"),
        ("revlink", "l", '', "Revision link (revlink)"),
        ("encoding", "e", 'utf8', "Encoding of other parameters"),
    ]

    buildbotOptions = [
        ['master', 'master'],
        ['who', 'who'],
        ['branch', 'branch'],
        ['category', 'category'],
        ['vc', 'vc'],
    ]

    requiredOptions = ['who', 'master']

    def getSynopsis(self):
        return "Usage:    buildbot sendchange [options] filenames.."

    def parseArgs(self, *args):
        self['files'] = args

    def opt_property(self, property):
        name, value = property.split(':', 1)
        self['properties'][name] = value

    def postOptions(self):
        super().postOptions()

        if self.get("revision_file"):
            with open(self["revision_file"], "r", encoding='utf-8') as f:
                self['revision'] = f.read()

        if self.get('when'):
            try:
                self['when'] = float(self['when'])
            except (TypeError, ValueError) as e:
                raise usage.UsageError(f"invalid 'when' value {self['when']}") from e
        else:
            self['when'] = None

        if not self.get('comments') and self.get('logfile'):
            if self['logfile'] == "-":
                self['comments'] = sys.stdin.read()
            else:
                with open(self['logfile'], "rt", encoding='utf-8') as f:
                    self['comments'] = f.read()
        if self.get('comments') is None:
            self['comments'] = ""

        # fix up the auth with a password if none was given
        auth = self.get('auth')
        if ':' not in auth:
            pw = getpass.getpass(f"Enter password for '{auth}': ")
            auth = f"{auth}:{pw}"
        self['auth'] = tuple(auth.split(':', 1))

        vcs = ['cvs', 'svn', 'darcs', 'hg', 'bzr', 'git', 'mtn', 'p4']
        if self.get('vc') and self.get('vc') not in vcs:
            raise usage.UsageError(f"vc must be one of {', '.join(vcs)}")

        validateMasterOption(self.get('master'))


class TryOptions(base.SubcommandOptions):
    subcommandFunction = "buildbot.scripts.trycmd.trycmd"
    optParameters = [
        ["connect", "c", None,
         "How to reach the buildmaster, either 'ssh' or 'pb'"],
        # for ssh, use --host, --username, --jobdir and optionally
        # --ssh
        ["host", None, None,
         "Hostname (used by ssh) for the buildmaster"],
        ["port", None, None,
         "Port (used by ssh) for the buildmaster"],
        ["jobdir", None, None,
         "Directory (on the buildmaster host) where try jobs are deposited"],
        ["ssh", None, None,
         "Command to use instead of the default \"ssh\""],
        ["username", "u", None,
         "Username performing the try build"],
        # for PB, use --master, --username, and --passwd
        ["master", "m", None,
         "Location of the buildmaster's Try server (host:port)"],
        ["passwd", None, None,
         "Password for PB authentication"],
        ["who", "w", None,
         "Who is responsible for the try build"],
        ["comment", "C", None,
         "A comment which can be used in notifications for this build"],

        # for ssh to accommodate running in a virtualenv on the buildmaster
        ["buildbotbin", None, "buildbot",
         "buildbot binary to use on the buildmaster host"],

        ["diff", None, None,
         "Filename of a patch to use instead of scanning a local tree. "
         "Use '-' for stdin."],
        ["patchlevel", "p", 0,
         "Number of slashes to remove from patch pathnames, "
         "like the -p option to 'patch'"],

        ["baserev", None, None,
         "Base revision to use instead of scanning a local tree."],

        ["vc", None, None,
         "The VC system in use, one of: bzr, cvs, darcs, git, hg, "
         "mtn, p4, svn"],
        ["branch", None, None,
         "The branch in use, for VC systems that can't figure it out "
         "themselves"],
        ["repository", None, None,
         "Repository to use, instead of path to working directory."],

        ["builder", "b", None,
         "Run the trial build on this Builder. Can be used multiple times."],
        ["properties", None, None,
         "A set of properties made available in the build environment, "
         "format is --properties=prop1=value1,prop2=value2,.. "
         "option can be specified multiple times."],
        ["property", None, None,
         "A property made available in the build environment, "
         "format:prop=value. Can be used multiple times."],

        ["topfile", None, None,
         "Name of a file at the top of the tree, used to find the top. "
         "Only needed for SVN and CVS."],
        ["topdir", None, None,
         "Path to the top of the working copy. Only needed for SVN and CVS."],
    ]

    optFlags = [
        ["wait", None,
         "wait until the builds have finished"],
        ["dryrun", 'n',
         "Gather info, but don't actually submit."],
        ["get-builder-names", None,
         "Get the names of available builders. Doesn't submit anything. "
         "Only supported for 'pb' connections."],
        ["quiet", "q",
         "Don't print status of current builds while waiting."],
    ]

    # Mapping of .buildbot/options names to command-line options
    buildbotOptions = [
        ['try_connect', 'connect'],
        # [ 'try_builders', 'builders' ], <-- handled in postOptions
        ['try_vc', 'vc'],
        ['try_branch', 'branch'],
        ['try_repository', 'repository'],
        ['try_topdir', 'topdir'],
        ['try_topfile', 'topfile'],
        ['try_host', 'host'],
        ['try_username', 'username'],
        ['try_jobdir', 'jobdir'],
        ['try_ssh', 'ssh'],
        ['try_buildbotbin', 'buildbotbin'],
        ['try_passwd', 'passwd'],
        ['try_master', 'master'],
        ['try_who', 'who'],
        ['try_comment', 'comment'],
        # [ 'try_wait', 'wait' ], <-- handled in postOptions
        # [ 'try_quiet', 'quiet' ], <-- handled in postOptions

        # Deprecated command mappings from the quirky old days:
        ['try_masterstatus', 'master'],
        ['try_dir', 'jobdir'],
        ['try_password', 'passwd'],
    ]

    def __init__(self):
        super().__init__()
        self['builders'] = []
        self['properties'] = {}

    def opt_builder(self, option):
        self['builders'].append(option)

    def opt_properties(self, option):
        # We need to split the value of this option
        # into a dictionary of properties
        propertylist = option.split(",")
        for prop in propertylist:
            splitproperty = prop.split("=", 1)
            self['properties'][splitproperty[0]] = splitproperty[1]

    def opt_property(self, option):
        name, _, value = option.partition("=")
        self['properties'][name] = value

    def opt_patchlevel(self, option):
        self['patchlevel'] = int(option)

    def getSynopsis(self):
        return "Usage:    buildbot try [options]"

    def postOptions(self):
        super().postOptions()
        opts = self.optionsFile
        if not self['builders']:
            self['builders'] = opts.get('try_builders', [])
        if opts.get('try_wait', False):
            self['wait'] = True
        if opts.get('try_quiet', False):
            self['quiet'] = True
        # get the global 'masterstatus' option if it's set and no master
        # was specified otherwise
        if not self['master']:
            self['master'] = opts.get('masterstatus', None)

        if self['connect'] == 'pb':
            if not self['master']:
                raise usage.UsageError("master location must be specified"
                                       "for 'pb' connections")
            validateMasterOption(self['master'])


class TryServerOptions(base.SubcommandOptions):
    subcommandFunction = "buildbot.scripts.tryserver.tryserver"
    optParameters = [
        ["jobdir", None, None, "the jobdir (maildir) for submitting jobs"],
    ]
    requiredOptions = ['jobdir']

    def getSynopsis(self):
        return "Usage:    buildbot tryserver [options]"

    def postOptions(self):
        if not self['jobdir']:
            raise usage.UsageError('jobdir is required')


class CheckConfigOptions(base.SubcommandOptions):
    subcommandFunction = "buildbot.scripts.checkconfig.checkconfig"
    optFlags = [
        ['quiet', 'q', "Don't display error messages or tracebacks"],
    ]

    # on tab completion, suggest files as first argument
    if hasattr(usage, 'Completions'):
        # only set completion suggestion if running with
        # twisted version (>=11.1.0) that supports it
        compData = usage.Completions(extraActions=[usage.CompleteFiles()])

    def getSynopsis(self):
        return "Usage:\t\tbuildbot checkconfig [configFile]\n" + \
            "\t\tIf not specified, the config file specified in " + \
            "'buildbot.tac' from the current directory will be used"

    def parseArgs(self, *args):
        if len(args) >= 1:
            self['configFile'] = args[0]


class UserOptions(base.SubcommandOptions):
    subcommandFunction = "buildbot.scripts.user.user"
    optParameters = [
        ["master", "m", None,
         "Location of the buildmaster's user service (host:port)"],
        ["username", "u", None,
         "Username for PB authentication"],
        ["passwd", "p", None,
         "Password for PB authentication"],
        ["op", None, None,
         "User management operation: add, remove, update, get"],
        ["bb_username", None, None,
         "Username to set for a given user. Only available on 'update', "
         "and bb_password must be given as well."],
        ["bb_password", None, None,
         "Password to set for a given user. Only available on 'update', "
         "and bb_username must be given as well."],
        ["ids", None, None,
         "User's identifiers, used to find users in 'remove' and 'get' "
         "Can be specified multiple times (--ids=id1,id2,id3)"],
        ["info", None, None,
         "User information in the form: --info=type=value,type=value,.. "
         "Used in 'add' and 'update', can be specified multiple times.  "
         "Note that 'update' requires --info=id:type=value..."]
    ]
    buildbotOptions = [
        ['master', 'master'],
        ['user_master', 'master'],
        ['user_username', 'username'],
        ['user_passwd', 'passwd'],
    ]
    requiredOptions = ['master']

    longdesc = textwrap.dedent("""
    Currently implemented types for --info= are:\n
    git, svn, hg, cvs, darcs, bzr, email
    """)

    def __init__(self):
        super().__init__()
        self['ids'] = []
        self['info'] = []

    def opt_ids(self, option):
        id_list = option.split(",")
        self['ids'].extend(id_list)

    def opt_info(self, option):
        # splits info into type/value dictionary, appends to info
        info_list = option.split(",")
        info_elem = {}

        if len(info_list) == 1 and '=' not in info_list[0]:
            info_elem["identifier"] = info_list[0]
            self['info'].append(info_elem)
        else:
            for info_item in info_list:
                split_info = info_item.split("=", 1)

                # pull identifier from update --info
                if ":" in split_info[0]:
                    split_id = split_info[0].split(":")
                    info_elem["identifier"] = split_id[0]
                    split_info[0] = split_id[1]

                info_elem[split_info[0]] = split_info[1]
            self['info'].append(info_elem)

    def getSynopsis(self):
        return "Usage:    buildbot user [options]"

    def _checkValidTypes(self, info):
        from buildbot.process.users import users
        valid = set(['identifier', 'email'] + users.srcs)

        for user in info:
            for attr_type in user:
                if attr_type not in valid:
                    raise usage.UsageError("Type not a valid attr_type, must be in: "
                                           f"{', '.join(valid)}")

    def postOptions(self):
        super().postOptions()

        validateMasterOption(self.get('master'))

        op = self.get('op')
        if not op:
            raise usage.UsageError("you must specify an operation: add, "
                                   "remove, update, get")
        if op not in ['add', 'remove', 'update', 'get']:
            raise usage.UsageError(f"bad op {repr(op)}, use 'add', 'remove', 'update', "
                                   "or 'get'")

        if not self.get('username') or not self.get('passwd'):
            raise usage.UsageError("A username and password must be given")

        bb_username = self.get('bb_username')
        bb_password = self.get('bb_password')
        if bb_username or bb_password:
            if op != 'update':
                raise usage.UsageError("bb_username and bb_password only work "
                                       "with update")
            if not bb_username or not bb_password:
                raise usage.UsageError("Must specify both bb_username and "
                                       "bb_password or neither.")

        info = self.get('info')
        ids = self.get('ids')

        # check for erroneous args
        if not info and not ids:
            raise usage.UsageError("must specify either --ids or --info")

        if op in ('add', 'update'):
            if ids:
                raise usage.UsageError("cannot use --ids with 'add' or "
                                       "'update'")
            self._checkValidTypes(info)
            if op == 'update':
                for user in info:
                    if 'identifier' not in user:
                        raise usage.UsageError("no ids found in update info; "
                                               "use: --info=id:type=value,type=value,..")
            if op == 'add':
                for user in info:
                    if 'identifier' in user:
                        raise usage.UsageError("identifier found in add info, "
                                               "use: --info=type=value,type=value,..")
        if op in ('remove', 'get'):
            if info:
                raise usage.UsageError("cannot use --info with 'remove' "
                                       "or 'get'")


class DataSpecOption(base.BasedirMixin, base.SubcommandOptions):
    subcommandFunction = "buildbot.scripts.dataspec.dataspec"
    optParameters = [
        ['out', 'o', "dataspec.json", "output to specified path"],
        ['global', 'g', None,
            "output a js script, that sets a global, for inclusion in testsuite"],
    ]

    def getSynopsis(self):
        return "Usage:   buildbot dataspec [options]"


class GenGraphQLOption(base.BasedirMixin, base.SubcommandOptions):
    subcommandFunction = "buildbot.scripts.gengraphql.gengraphql"
    optParameters = [
        ['out', 'o', "graphql.schema", "output to specified path"],
    ]

    def getSynopsis(self):
        return "Usage:   buildbot graphql-schema [options]"


class DevProxyOptions(base.BasedirMixin, base.SubcommandOptions):

    """Run a fake web server serving the local ui frontend and a distant rest and websocket api.
    This command required aiohttp to be installed in the virtualenv"""

    subcommandFunction = "buildbot.scripts.devproxy.devproxy"
    optFlags = [
        ["unsafe_ssl", None, "Bypass ssl certificate validation"],
    ]
    optParameters = [
        ["port", "p", 8011,
         "http port to use"],
        ["plugins", None, None,
         "plugin config to use. As json string e.g: "
         "--plugins='{\"custom_plugin\": {\"option1\": true}}'"],
        ["auth_cookie", None, None,
         "TWISTED_SESSION cookie to be used for auth "
         "(taken in developer console: in document.cookie variable)"],
        ["buildbot_url", "b", "https://buildbot.buildbot.net",
         "real buildbot url to proxy to (can be http or https)"]
    ]


class CleanupDBOptions(base.BasedirMixin, base.SubcommandOptions):
    subcommandFunction = "buildbot.scripts.cleanupdb.cleanupDatabase"
    optFlags = [
        ["quiet", "q", "Do not emit the commands being run"],
        ["force", "f",
            "Force log recompression (useful when changing compression algorithm)"],
        # when this command has several maintenance jobs, we should make
        # them optional here. For now there is only one.
    ]
    optParameters = [
    ]

    def getSynopsis(self):
        return "Usage:    buildbot cleanupdb [options] [<basedir>]"

    longdesc = textwrap.dedent("""
    This command takes an existing buildmaster working directory and
    do some optimization on the database.

    This command is frontend for various database maintenance jobs:

    - optimiselogs: This optimization groups logs into bigger chunks
      to apply higher level of compression.

    This command uses the database specified in
    the master configuration file.  If you wish to use a database other than
    the default (sqlite), be sure to set that parameter before upgrading.
    """)


class Options(usage.Options):
    synopsis = "Usage:    buildbot <command> [command options]"

    subCommands = [
        ['create-master', None, CreateMasterOptions,
         "Create and populate a directory for a new buildmaster"],
        ['upgrade-master', None, UpgradeMasterOptions,
         "Upgrade an existing buildmaster directory for the current version"],
        ['start', None, StartOptions,
         "Start a buildmaster"],
        ['stop', None, StopOptions,
         "Stop a buildmaster"],
        ['restart', None, RestartOptions,
         "Restart a buildmaster"],
        ['reconfig', None, ReconfigOptions,
         "SIGHUP a buildmaster to make it re-read the config file"],
        ['sighup', None, ReconfigOptions,
         "SIGHUP a buildmaster to make it re-read the config file"],
        ['sendchange', None, SendChangeOptions,
         "Send a change to the buildmaster"],
        ['try', None, TryOptions,
         "Run a build with your local changes"],
        ['tryserver', None, TryServerOptions,
         "buildmaster-side 'try' support function, not for users"],
        ['checkconfig', None, CheckConfigOptions,
         "test the validity of a master.cfg config file"],
        ['user', None, UserOptions,
         "Manage users in buildbot's database"],
        ['dataspec', None, DataSpecOption,
         "Output data api spec"],
        ['dev-proxy', None, DevProxyOptions,
         "Run a fake web server serving the local ui frontend and a distant rest and websocket api."
         ],
        ['graphql-schema', None, GenGraphQLOption,
         "Output graphql api schema"],
        ['cleanupdb', None, CleanupDBOptions,
         "cleanup the database"
         ]
    ]

    def opt_version(self):
        print(f"Buildbot version: {buildbot.version}")
        super().opt_version()

    def opt_verbose(self):
        from twisted.python import log
        log.startLogging(sys.stderr)

    def postOptions(self):
        if not hasattr(self, 'subOptions'):
            raise usage.UsageError("must specify a command")


def run():
    config = Options()
    check_functional_environment(buildbot.config)
    try:
        config.parseOptions(sys.argv[1:])
    except usage.error as e:
        print(f"{sys.argv[0]}:  {e}")
        print()

        c = getattr(config, 'subOptions', config)
        print(str(c))
        sys.exit(1)

    subconfig = config.subOptions
    subcommandFunction = reflect.namedObject(subconfig.subcommandFunction)
    sys.exit(subcommandFunction(subconfig))



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import absolute_import
from __future__ import print_function

import io
import os
import shutil
import sys
import tarfile

from twisted.internet import defer
from twisted.internet import reactor
from twisted.python import failure
from twisted.python import runtime
from twisted.trial import unittest

from buildbot_worker.commands import transfer
from buildbot_worker.test.fake.remote import FakeRemote
from buildbot_worker.test.util.command import CommandTestMixin


class FakeMasterMethods(object):
    # a fake to represent any of:
    # - FileWriter
    # - FileDirectoryWriter
    # - FileReader

    def __init__(self, add_update):
        self.add_update = add_update

        self.delay_write = False
        self.count_writes = False
        self.keep_data = False
        self.write_out_of_space_at = None

        self.delay_read = False
        self.count_reads = False

        self.unpack_fail = False

        self.written = False
        self.read = False
        self.data = b''

    def remote_write(self, data):
        if self.write_out_of_space_at is not None:
            self.write_out_of_space_at -= len(data)
            if self.write_out_of_space_at <= 0:
                f = failure.Failure(RuntimeError("out of space"))
                return defer.fail(f)
        if self.count_writes:
            self.add_update('write {0}'.format(len(data)))
        elif not self.written:
            self.add_update('write(s)')
            self.written = True

        if self.keep_data:
            self.data += data

        if self.delay_write:
            d = defer.Deferred()
            reactor.callLater(0.01, d.callback, None)
            return d
        return None

    def remote_read(self, length):
        if self.count_reads:
            self.add_update('read {0}'.format(length))
        elif not self.read:
            self.add_update('read(s)')
            self.read = True

        if not self.data:
            return ''

        _slice, self.data = self.data[:length], self.data[length:]
        if self.delay_read:
            d = defer.Deferred()
            reactor.callLater(0.01, d.callback, _slice)
            return d
        return _slice

    def remote_unpack(self):
        self.add_update('unpack')
        if self.unpack_fail:
            return defer.fail(failure.Failure(RuntimeError("out of space")))
        return None

    def remote_utime(self, accessed_modified):
        self.add_update('utime - {0}'.format(accessed_modified[0]))

    def remote_close(self):
        self.add_update('close')


class TestUploadFile(CommandTestMixin, unittest.TestCase):

    def setUp(self):
        self.setUpCommand()

        self.fakemaster = FakeMasterMethods(self.add_update)

        # write 180 bytes of data to upload
        self.datadir = os.path.join(self.basedir, 'workdir')
        if os.path.exists(self.datadir):
            shutil.rmtree(self.datadir)
        os.makedirs(self.datadir)

        self.datafile = os.path.join(self.datadir, 'data')
        # note: use of 'wb' here ensures newlines aren't translated on the
        # upload
        with open(self.datafile, mode="wb") as f:
            f.write(b"this is some data\n" * 10)

    def tearDown(self):
        self.tearDownCommand()

        if os.path.exists(self.datadir):
            shutil.rmtree(self.datadir)

    @defer.inlineCallbacks
    def test_simple(self):
        self.fakemaster.count_writes = True    # get actual byte counts

        path = os.path.join(self.basedir, 'workdir', os.path.expanduser('data'))
        self.make_command(transfer.WorkerFileUploadCommand, {
            'path': path,
            'writer': FakeRemote(self.fakemaster),
            'maxsize': 1000,
            'blocksize': 64,
            'keepstamp': False
        })

        yield self.run_command()

        self.assertUpdates([
            ('header', 'sending {0}\n'.format(self.datafile)),
            'write 64', 'write 64', 'write 52', 'close',
            ('rc', 0)
        ])

    @defer.inlineCallbacks
    def test_truncated(self):
        self.fakemaster.count_writes = True    # get actual byte counts

        path = os.path.join(self.basedir, 'workdir', os.path.expanduser('data'))
        self.make_command(transfer.WorkerFileUploadCommand, {
            'path': path,
            'writer': FakeRemote(self.fakemaster),
            'maxsize': 100,
            'blocksize': 64,
            'keepstamp': False
        })

        yield self.run_command()

        self.assertUpdates([
            ('header', 'sending {0}\n'.format(self.datafile)),
            'write 64', 'write 36', 'close',
            ('rc', 1),
            ('stderr', "Maximum filesize reached, truncating file '{0}'".format(self.datafile))
        ])

    @defer.inlineCallbacks
    def test_missing(self):
        path = os.path.join(self.basedir, 'workdir', os.path.expanduser('data-nosuch'))
        self.make_command(transfer.WorkerFileUploadCommand, {
            'path': path,
            'writer': FakeRemote(self.fakemaster),
            'maxsize': 100,
            'blocksize': 64,
            'keepstamp': False
        })

        yield self.run_command()

        df = self.datafile + "-nosuch"
        self.assertUpdates([
            ('header', 'sending {0}\n'.format(df)),
            'close',
            ('rc', 1),
            ('stderr', "Cannot open file '{0}' for upload".format(df))
        ])

    @defer.inlineCallbacks
    def test_out_of_space(self):
        self.fakemaster.write_out_of_space_at = 70
        self.fakemaster.count_writes = True    # get actual byte counts

        path = os.path.join(self.basedir, 'workdir', os.path.expanduser('data'))
        self.make_command(transfer.WorkerFileUploadCommand, {
            'path': path,
            'writer': FakeRemote(self.fakemaster),
            'maxsize': 1000,
            'blocksize': 64,
            'keepstamp': False
        })

        yield self.assertFailure(self.run_command(), RuntimeError)

        self.assertUpdates([
            ('header', 'sending {0}\n'.format(self.datafile)),
            'write 64', 'close',
            ('rc', 1)
        ])

    @defer.inlineCallbacks
    def test_interrupted(self):
        self.fakemaster.delay_write = True  # write very slowly

        path = os.path.join(self.basedir, 'workdir', os.path.expanduser('data'))
        self.make_command(transfer.WorkerFileUploadCommand, {
            'path': path,
            'writer': FakeRemote(self.fakemaster),
            'maxsize': 100,
            'blocksize': 2,
            'keepstamp': False
        })

        d = self.run_command()

        # wait a jiffy..
        interrupt_d = defer.Deferred()
        reactor.callLater(0.01, interrupt_d.callback, None)

        # and then interrupt the step
        def do_interrupt(_):
            return self.cmd.interrupt()
        interrupt_d.addCallback(do_interrupt)

        yield defer.DeferredList([d, interrupt_d])

        self.assertUpdates([
            ('header', 'sending {0}\n'.format(self.datafile)), 'write(s)', 'close', ('rc', 1)
        ])

    @defer.inlineCallbacks
    def test_timestamp(self):
        self.fakemaster.count_writes = True    # get actual byte counts
        timestamp = (os.path.getatime(self.datafile),
                     os.path.getmtime(self.datafile))

        path = os.path.join(self.basedir, 'workdir', os.path.expanduser('data'))
        self.make_command(transfer.WorkerFileUploadCommand, {
            'path': path,
            'writer': FakeRemote(self.fakemaster),
            'maxsize': 1000,
            'blocksize': 64,
            'keepstamp': True,
        })

        yield self.run_command()

        self.assertUpdates([
            ('header', 'sending {0}\n'.format(self.datafile)),
            'write 64', 'write 64', 'write 52',
            'close', 'utime - {0}'.format(timestamp[0]),
            ('rc', 0)
        ])


class TestWorkerDirectoryUpload(CommandTestMixin, unittest.TestCase):

    def setUp(self):
        self.setUpCommand()

        self.fakemaster = FakeMasterMethods(self.add_update)

        # write a directory to upload
        self.datadir = os.path.join(self.basedir, 'workdir', 'data')
        if os.path.exists(self.datadir):
            shutil.rmtree(self.datadir)
        os.makedirs(self.datadir)
        with open(os.path.join(self.datadir, "aa"), mode="wb") as f:
            f.write(b"lots of a" * 100)
        with open(os.path.join(self.datadir, "bb"), mode="wb") as f:
            f.write(b"and a little b" * 17)

    def tearDown(self):
        self.tearDownCommand()

        if os.path.exists(self.datadir):
            shutil.rmtree(self.datadir)

    @defer.inlineCallbacks
    def test_simple(self, compress=None):
        self.fakemaster.keep_data = True
        path = os.path.join(self.basedir, 'workdir', os.path.expanduser('data'))
        self.make_command(transfer.WorkerDirectoryUploadCommand, {
            'workdir': 'workdir',
            'path': path,
            'writer': FakeRemote(self.fakemaster),
            'maxsize': None,
            'blocksize': 512,
            'compress': compress,
        })

        yield self.run_command()

        self.assertUpdates([
            ('header', 'sending {0}\n'.format(self.datadir)),
            'write(s)', 'unpack',  # note no 'close"
            ('rc', 0)
        ])

        f = io.BytesIO(self.fakemaster.data)
        a = tarfile.open(fileobj=f, name='check.tar', mode="r")
        exp_names = ['.', 'aa', 'bb']
        got_names = [n.rstrip('/') for n in a.getnames()]
        # py27 uses '' instead of '.'
        got_names = sorted([n or '.' for n in got_names])
        self.assertEqual(got_names, exp_names, "expected archive contents")
        a.close()
        f.close()

    # try it again with bz2 and gzip
    def test_simple_bz2(self):
        return self.test_simple('bz2')

    def test_simple_gz(self):
        return self.test_simple('gz')

    # except bz2 can't operate in stream mode on py24
    if sys.version_info[:2] <= (2, 4):
        test_simple_bz2.skip = "bz2 stream decompression not supported on Python-2.4"

    @defer.inlineCallbacks
    def test_out_of_space_unpack(self):
        self.fakemaster.keep_data = True
        self.fakemaster.unpack_fail = True
        path = os.path.join(self.basedir, 'workdir', os.path.expanduser('data'))

        self.make_command(transfer.WorkerDirectoryUploadCommand, {
            'path': path,
            'workersrc': 'data',
            'writer': FakeRemote(self.fakemaster),
            'maxsize': None,
            'blocksize': 512,
            'compress': None
        })

        yield self.assertFailure(self.run_command(), RuntimeError)

        self.assertUpdates([
            ('header', 'sending {0}\n'.format(self.datadir)),
            'write(s)', 'unpack',
            ('rc', 1)
        ])


class TestDownloadFile(CommandTestMixin, unittest.TestCase):

    def setUp(self):
        self.setUpCommand()

        self.fakemaster = FakeMasterMethods(self.add_update)

        # the command will write to the basedir, so make sure it exists
        if os.path.exists(self.basedir):
            shutil.rmtree(self.basedir)
        os.makedirs(self.basedir)

    def tearDown(self):
        self.tearDownCommand()

        if os.path.exists(self.basedir):
            shutil.rmtree(self.basedir)

    @defer.inlineCallbacks
    def test_simple(self):
        self.fakemaster.count_reads = True    # get actual byte counts
        self.fakemaster.data = test_data = b'1234' * 13
        assert len(self.fakemaster.data) == 52

        path = os.path.join(self.basedir, os.path.expanduser('data'))
        self.make_command(transfer.WorkerFileDownloadCommand, {
            'path': path,
            'reader': FakeRemote(self.fakemaster),
            'maxsize': None,
            'blocksize': 32,
            'mode': 0o777
        })

        yield self.run_command()

        self.assertUpdates([
            'read 32', 'read 32', 'read 32', 'close',
            ('rc', 0)
        ])
        datafile = os.path.join(self.basedir, 'data')
        self.assertTrue(os.path.exists(datafile))
        with open(datafile, mode="rb") as f:
            datafileContent = f.read()
        self.assertEqual(datafileContent, test_data)
        if runtime.platformType != 'win32':
            self.assertEqual(os.stat(datafile).st_mode & 0o777, 0o777)

    @defer.inlineCallbacks
    def test_mkdir(self):
        self.fakemaster.data = test_data = b'hi'

        path = os.path.join(self.basedir, 'workdir',
                            os.path.expanduser(os.path.join('subdir', 'data')))
        self.make_command(transfer.WorkerFileDownloadCommand, {
            'path': path,
            'reader': FakeRemote(self.fakemaster),
            'maxsize': None,
            'blocksize': 32,
            'mode': 0o777
        })

        yield self.run_command()

        self.assertUpdates([
            'read(s)', 'close',
            ('rc', 0)
        ])
        datafile = os.path.join(self.basedir, 'workdir', 'subdir', 'data')
        self.assertTrue(os.path.exists(datafile))
        with open(datafile, mode="rb") as f:
            datafileContent = f.read()
        self.assertEqual(datafileContent, test_data)

    @defer.inlineCallbacks
    def test_failure(self):
        self.fakemaster.data = 'hi'

        os.makedirs(os.path.join(self.basedir, 'dir'))
        path = os.path.join(self.basedir, os.path.expanduser('dir'))
        self.make_command(transfer.WorkerFileDownloadCommand, {
            'path': path,
            'reader': FakeRemote(self.fakemaster),
            'maxsize': None,
            'blocksize': 32,
            'mode': 0o777
        })

        yield self.run_command()

        self.assertUpdates([
            'close',
            ('rc', 1),
            ('stderr', "Cannot open file '{0}' for download".format(
             os.path.join(self.basedir, 'dir')))
        ])

    @defer.inlineCallbacks
    def test_truncated(self):
        self.fakemaster.data = test_data = b'tenchars--' * 10

        path = os.path.join(self.basedir, os.path.expanduser('data'))
        self.make_command(transfer.WorkerFileDownloadCommand, {
            'path': path,
            'reader': FakeRemote(self.fakemaster),
            'maxsize': 50,
            'blocksize': 32,
            'mode': 0o777
        })
        yield self.run_command()

        self.assertUpdates([
            'read(s)', 'close',
            ('rc', 1),
            ('stderr', "Maximum filesize reached, truncating file '{0}'".format(
             os.path.join(self.basedir, 'data')))
        ])
        datafile = os.path.join(self.basedir, 'data')
        self.assertTrue(os.path.exists(datafile))
        with open(datafile, mode="rb") as f:
            data = f.read()
        self.assertEqual(data, test_data[:50])

    @defer.inlineCallbacks
    def test_interrupted(self):
        self.fakemaster.data = b'tenchars--' * 100  # 1k
        self.fakemaster.delay_read = True  # read very slowly

        path = os.path.join(self.basedir, os.path.expanduser('data'))
        self.make_command(transfer.WorkerFileDownloadCommand, {
            'path': path,
            'reader': FakeRemote(self.fakemaster),
            'maxsize': 100,
            'blocksize': 2,
            'mode': 0o777
        })

        d = self.run_command()

        # wait a jiffy..
        interrupt_d = defer.Deferred()
        reactor.callLater(0.01, interrupt_d.callback, None)

        # and then interrupt the step
        def do_interrupt(_):
            return self.cmd.interrupt()
        interrupt_d.addCallback(do_interrupt)

        yield defer.DeferredList([d, interrupt_d])

        self.assertUpdates([
            'read(s)', 'close', ('rc', 1)
        ])



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from twisted.application import strports
from twisted.internet import defer
from twisted.python import log

from buildbot.util import service


class BaseManager(service.AsyncMultiService):
    """
    A centralized manager for connection ports and authentication on them.
    Allows various pieces of code to request a (port, username) combo, along
    with a password and a connection factory.
    """
    def __init__(self, name):
        super().__init__()
        self.setName(name)
        self.dispatchers = {}

    @defer.inlineCallbacks
    def register(self, config_portstr, username, password, pfactory):
        """
        Register a connection code to be executed after a user with its USERNAME/PASSWORD
        was authenticated and a valid high level connection can be established on a PORTSTR.
        Returns a Registration object which can be used to unregister later.
        """
        portstr = config_portstr

        # do some basic normalization of portstrs
        if isinstance(portstr, int) or ':' not in portstr:
            portstr = f"tcp:{portstr}".format(portstr)

        reg = Registration(self, portstr, username)

        if portstr not in self.dispatchers:
            disp = self.dispatchers[portstr] = self.dispatcher_class(config_portstr, portstr)
            yield disp.setServiceParent(self)
        else:
            disp = self.dispatchers[portstr]

        disp.register(username, password, pfactory)

        return reg

    @defer.inlineCallbacks
    def _unregister(self, registration):
        disp = self.dispatchers[registration.portstr]
        disp.unregister(registration.username)
        registration.username = None
        if not disp.users:
            del self.dispatchers[registration.portstr]
            yield disp.disownServiceParent()


class Registration:

    def __init__(self, manager, portstr, username):
        self.portstr = portstr
        "portstr this registration is active on"
        self.username = username
        "username of this registration"
        self.manager = manager

    def __repr__(self):
        return f"<base.Registration for {self.username} on {self.portstr}>"

    def unregister(self):
        """
        Unregister this registration, removing the username from the port, and
        closing the port if there are no more users left.  Returns a Deferred.
        """
        return self.manager._unregister(self)

    def getPort(self):
        """
        Helper method for testing; returns the TCP port used for this
        registration, even if it was specified as 0 and thus allocated by the
        OS.
        """
        disp = self.manager.dispatchers[self.portstr]
        return disp.port.getHost().port


class BaseDispatcher(service.AsyncService):
    debug = False

    def __init__(self, portstr):
        self.portstr = portstr
        self.users = {}
        self.port = None

    def __repr__(self):
        return f'<base.BaseDispatcher for {", ".join(list(self.users))} on {self.portstr}>'

    def start_listening_port(self):
        return strports.listen(self.portstr, self.serverFactory)

    def startService(self):
        assert not self.port
        self.port = self.start_listening_port()

        return super().startService()

    @defer.inlineCallbacks
    def stopService(self):
        # stop listening on the port when shut down
        assert self.port
        port, self.port = self.port, None
        yield port.stopListening()
        yield super().stopService()

    def register(self, username, password, pfactory):
        if self.debug:
            log.msg(f"registering username '{username}' on port {self.portstr}: {pfactory}")
        if username in self.users:
            raise KeyError(f"username '{username}' is already registered on port {self.portstr}")
        self.users[username] = (password, pfactory)

    def unregister(self, username):
        if self.debug:
            log.msg(f"unregistering username '{username}' on port {self.portstr}")
        del self.users[username]



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from twisted.internet import defer
from twisted.python import log
from twisted.trial import unittest

from buildbot.test.util.logging import LoggingMixin
from buildbot.util.async_sort import async_sort


class AsyncSort(unittest.TestCase, LoggingMixin):

    def setUp(self) -> None:
        self.setUpLogging()
        return super().setUp()

    @defer.inlineCallbacks
    def test_sync_call(self):
        l = ["b", "c", "a"]
        yield async_sort(l, lambda x: x)
        return self.assertEqual(l, ["a", "b", "c"])

    @defer.inlineCallbacks
    def test_async_call(self):
        l = ["b", "c", "a"]
        yield async_sort(l, defer.succeed)
        self.assertEqual(l, ["a", "b", "c"])

    @defer.inlineCallbacks
    def test_async_fail(self):
        l = ["b", "c", "a"]
        self.patch(log, "err", lambda f: None)

        class SortFail(Exception):
            pass
        with self.assertRaises(SortFail):
            yield async_sort(l, lambda x:
                defer.succeed(x) if x != "a" else defer.fail(SortFail("ono")))

        self.assertEqual(len(self.flushLoggedErrors(SortFail)), 1)
        self.assertEqual(l, ["b", "c", "a"])



from twisted.internet import defer

from buildbot.schedulers.forcesched import ChoiceStringParameter
from buildbot.schedulers.forcesched import NestedParameter
from buildbot.schedulers.forcesched import StringParameter
from buildbot.schedulers.forcesched import ValidationError
from buildbot.www.plugin import Application

from .api import Api


class NestedExample(NestedParameter):

    """UI zone"""
    type = "nestedexample"
    PIZZA = "pizza"
    INGREDIENTS = "ingredients"

    def __init__(self, **kw):
        pizzaInput = StringParameter(label="type the name of your pizza",
                                     name=self.PIZZA,
                                     required=True)
        ingredientsInput = ChoiceStringParameter(name=self.INGREDIENTS,
                                                 label="ingredients necessary to make the pizza",
                                                 multiple=True,
                                                 strict=False,
                                                 default="",
                                                 choices=[])
        self.params = {self.PIZZA: pizzaInput,
                       self.INGREDIENTS: ingredientsInput}
        self.allIngredients = set(sum([ingr for ingr in Api.pizzaIngredients.values()],
                                      []))
        fields = self.params.values()
        super(NestedExample, self).__init__(self.type, label='', fields=fields, **kw)

    def createNestedPropertyName(self, propertyName):
        return "{}_{}".format(self.type, propertyName)

    @defer.inlineCallbacks
    def validateProperties(self, collector, properties):
        # we implement the check between the input and
        # the ingredients
        if properties[self.INGREDIENTS] not in self.allIngredients or\
           not properties[self.PIZZA]:
            # we trigger a specific error message in PIZZA only
            def f():
                return defer.fail(ValidationError('Invalid pizza'))
            nestedProp = self.createNestedPropertyName(self.PIZZA)
            yield collector.collectValidationErrors(nestedProp, f)

    @defer.inlineCallbacks
    def updateFromKwargs(self, kwargs, properties, collector, **kw):
        yield super(NestedExample, self).updateFromKwargs(kwargs, properties, collector, **kw)
        # the properties we have are in the form
        # {nestedexample: {input: <url>,
        #                 ingredients: <ingredients>}}
        # we just flatten the dict to have
        # - input, and
        # - ingredients
        # in properties
        for prop, val in properties.pop(self.type).items():
            properties[prop] = val
        yield self.validateProperties(collector, properties)


# create the interface for the setuptools entry point
ep = Application(__name__, "Buildbot nested parameter example")
api = Api(ep)
ep.resource.putChild("api", api.app.resource())



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Portions Copyright Buildbot Team Members

from twisted.internet import defer
from zope.interface import implementer

from buildbot import interfaces
from buildbot.util import service


@implementer(interfaces.IMachine)
class Machine(service.BuildbotService):

    def checkConfig(self, name, **kwargs):
        super().checkConfig(**kwargs)
        self.name = name
        self.workers = []

    @defer.inlineCallbacks
    def reconfigService(self, name, **kwargs):
        yield super().reconfigService(**kwargs)
        assert self.name == name

    def registerWorker(self, worker):
        assert worker.machine_name == self.name
        self.workers.append(worker)

    def unregisterWorker(self, worker):
        assert worker in self.workers
        self.workers.remove(worker)

    def __repr__(self):
        return f"<Machine '{self.name}' at {id(self)}>"



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from buildbot.schedulers.basic import AnyBranchScheduler
from buildbot.schedulers.basic import Scheduler
from buildbot.schedulers.dependent import Dependent
from buildbot.schedulers.timed import Nightly
from buildbot.schedulers.timed import Periodic
from buildbot.schedulers.triggerable import Triggerable
from buildbot.schedulers.trysched import Try_Jobdir
from buildbot.schedulers.trysched import Try_Userpass

_hush_pyflakes = [Scheduler, AnyBranchScheduler, Dependent,
                  Periodic, Nightly, Triggerable, Try_Jobdir, Try_Userpass]
del _hush_pyflakes



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import copy

from twisted.internet import defer
from twisted.python import log

from buildbot.data import base
from buildbot.data import sourcestamps as sourcestampsapi
from buildbot.data import types
from buildbot.db.buildsets import AlreadyCompleteError
from buildbot.process.buildrequest import BuildRequestCollapser
from buildbot.process.results import SUCCESS
from buildbot.process.results import worst_status
from buildbot.util import datetime2epoch
from buildbot.util import epoch2datetime


class Db2DataMixin:

    @defer.inlineCallbacks
    def db2data(self, bsdict):
        if not bsdict:
            return None

        buildset = bsdict.copy()

        # gather the actual sourcestamps, in parallel
        sourcestamps = []

        @defer.inlineCallbacks
        def getSs(ssid):
            ss = yield self.master.data.get(('sourcestamps', str(ssid)))
            sourcestamps.append(ss)
        yield defer.DeferredList([getSs(id)
                                  for id in buildset['sourcestamps']],
                                 fireOnOneErrback=True, consumeErrors=True)
        buildset['sourcestamps'] = sourcestamps

        # minor modifications
        buildset['submitted_at'] = datetime2epoch(buildset['submitted_at'])
        buildset['complete_at'] = datetime2epoch(buildset['complete_at'])

        return buildset

    fieldMapping = {
        'bsid': 'buildsets.id',
        'external_idstring': 'buildsets.external_idstring',
        'reason': 'buildsets.reason',
        'submitted_at': 'buildsets.submitted_at',
        'complete': 'buildsets.complete',
        'complete_at': 'buildsets.complete_at',
        'results': 'buildsets.results',
        'parent_buildid': 'buildsets.parent_buildid',
        'parent_relationship': 'buildsets.parent_relationship'
    }


class BuildsetEndpoint(Db2DataMixin, base.Endpoint):

    isCollection = False
    pathPatterns = """
        /buildsets/n:bsid
    """

    @defer.inlineCallbacks
    def get(self, resultSpec, kwargs):
        res = yield self.master.db.buildsets.getBuildset(kwargs['bsid'])
        res = yield self.db2data(res)
        return res


class BuildsetsEndpoint(Db2DataMixin, base.Endpoint):

    isCollection = True
    pathPatterns = """
        /buildsets
    """
    rootLinkName = 'buildsets'

    def get(self, resultSpec, kwargs):
        complete = resultSpec.popBooleanFilter('complete')
        resultSpec.fieldMapping = self.fieldMapping
        d = self.master.db.buildsets.getBuildsets(
            complete=complete, resultSpec=resultSpec)

        @d.addCallback
        def db2data(buildsets):
            d = defer.DeferredList([self.db2data(bs) for bs in buildsets],
                                   fireOnOneErrback=True, consumeErrors=True)

            @d.addCallback
            def getResults(res):
                return [r[1] for r in res]
            return d
        return d


class Buildset(base.ResourceType):

    name = "buildset"
    plural = "buildsets"
    endpoints = [BuildsetEndpoint, BuildsetsEndpoint]
    keyField = 'bsid'
    eventPathPatterns = """
        /buildsets/:bsid
    """

    class EntityType(types.Entity):
        bsid = types.Integer()
        external_idstring = types.NoneOk(types.String())
        reason = types.String()
        submitted_at = types.Integer()
        complete = types.Boolean()
        complete_at = types.NoneOk(types.Integer())
        results = types.NoneOk(types.Integer())
        sourcestamps = types.List(
            of=sourcestampsapi.SourceStamp.entityType)
        parent_buildid = types.NoneOk(types.Integer())
        parent_relationship = types.NoneOk(types.String())
    entityType = EntityType(name, 'Buildset')
    subresources = ["Property"]

    @base.updateMethod
    @defer.inlineCallbacks
    def addBuildset(self, waited_for, scheduler=None, sourcestamps=None, reason='',
                    properties=None, builderids=None, external_idstring=None,
                    parent_buildid=None, parent_relationship=None):
        if sourcestamps is None:
            sourcestamps = []
        if properties is None:
            properties = {}
        if builderids is None:
            builderids = []
        submitted_at = int(self.master.reactor.seconds())
        bsid, brids = yield self.master.db.buildsets.addBuildset(
            sourcestamps=sourcestamps, reason=reason,
            properties=properties, builderids=builderids,
            waited_for=waited_for, external_idstring=external_idstring,
            submitted_at=epoch2datetime(submitted_at),
            parent_buildid=parent_buildid, parent_relationship=parent_relationship)

        yield BuildRequestCollapser(self.master, list(brids.values())).collapse()

        # get each of the sourcestamps for this buildset (sequentially)
        bsdict = yield self.master.db.buildsets.getBuildset(bsid)
        sourcestamps = []
        for ssid in bsdict['sourcestamps']:
            sourcestamps.append(
                (yield self.master.data.get(('sourcestamps', str(ssid)))).copy()
            )

        # notify about the component build requests
        brResource = self.master.data.getResourceType("buildrequest")
        brResource.generateEvent(list(brids.values()), 'new')

        # and the buildset itself
        msg = dict(
            bsid=bsid,
            external_idstring=external_idstring,
            reason=reason,
            submitted_at=submitted_at,
            complete=False,
            complete_at=None,
            results=None,
            scheduler=scheduler,
            sourcestamps=sourcestamps)
        # TODO: properties=properties)
        self.produceEvent(msg, "new")

        log.msg(f"added buildset {bsid} to database")

        # if there are no builders, then this is done already, so send the
        # appropriate messages for that
        if not builderids:
            yield self.maybeBuildsetComplete(bsid)

        return (bsid, brids)

    @base.updateMethod
    @defer.inlineCallbacks
    def maybeBuildsetComplete(self, bsid):
        brdicts = yield self.master.db.buildrequests.getBuildRequests(
            bsid=bsid, complete=False)

        # if there are incomplete buildrequests, bail out
        if brdicts:
            return

        brdicts = yield self.master.db.buildrequests.getBuildRequests(bsid=bsid)

        # figure out the overall results of the buildset:
        cumulative_results = SUCCESS
        for brdict in brdicts:
            cumulative_results = worst_status(
                cumulative_results, brdict['results'])

        # get a copy of the buildset
        bsdict = yield self.master.db.buildsets.getBuildset(bsid)

        # if it's already completed, we're late to the game, and there's
        # nothing to do.
        #
        # NOTE: there's still a strong possibility of a race condition here,
        # which would cause buildset being completed twice.
        # in this case, the db layer will detect that and raise AlreadyCompleteError
        if bsdict['complete']:
            return

        # mark it as completed in the database
        complete_at = epoch2datetime(int(self.master.reactor.seconds()))
        try:
            yield self.master.db.buildsets.completeBuildset(bsid, cumulative_results,
                                                            complete_at=complete_at)
        except AlreadyCompleteError:
            return
        # get the sourcestamps for the message
        # get each of the sourcestamps for this buildset (sequentially)
        bsdict = yield self.master.db.buildsets.getBuildset(bsid)
        sourcestamps = []
        for ssid in bsdict['sourcestamps']:
            sourcestamps.append(
                copy.deepcopy(
                    (yield self.master.data.get(('sourcestamps', str(ssid))))
                )
            )

        msg = dict(
            bsid=bsid,
            external_idstring=bsdict['external_idstring'],
            reason=bsdict['reason'],
            sourcestamps=sourcestamps,
            submitted_at=bsdict['submitted_at'],
            complete=True,
            complete_at=complete_at,
            results=cumulative_results)
        # TODO: properties=properties)
        self.produceEvent(msg, "complete")



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import time

from twisted.internet import defer

from buildbot.util import asyncSleep


class BackoffTimeoutExceededError(Exception):
    pass


class ExponentialBackoffEngine:
    def __init__(self, start_seconds, multiplier, max_wait_seconds):
        if start_seconds < 0:
            raise ValueError('start_seconds cannot be negative')
        if multiplier < 0:
            raise ValueError('multiplier cannot be negative')
        if max_wait_seconds < 0:
            raise ValueError('max_wait_seconds cannot be negative')

        self.start_seconds = start_seconds
        self.multiplier = multiplier
        self.max_wait_seconds = max_wait_seconds

        self.on_success()

    def on_success(self):
        self.current_total_wait_seconds = 0
        self.current_wait_seconds = self.start_seconds

    def wait_on_failure(self):
        raise NotImplementedError()

    def calculate_wait_on_failure_seconds(self):
        if self.current_total_wait_seconds >= self.max_wait_seconds:
            raise BackoffTimeoutExceededError()

        seconds = self.current_wait_seconds
        self.current_wait_seconds *= self.multiplier
        if self.current_total_wait_seconds + seconds < self.max_wait_seconds:
            self.current_total_wait_seconds += seconds
        else:
            seconds = self.max_wait_seconds - self.current_total_wait_seconds
            self.current_total_wait_seconds = self.max_wait_seconds
        return seconds


class ExponentialBackoffEngineSync(ExponentialBackoffEngine):
    def wait_on_failure(self):
        seconds = self.calculate_wait_on_failure_seconds()
        time.sleep(seconds)


class ExponentialBackoffEngineAsync(ExponentialBackoffEngine):
    def __init__(self, reactor, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.reactor = reactor

    @defer.inlineCallbacks
    def wait_on_failure(self):
        seconds = self.calculate_wait_on_failure_seconds()
        yield asyncSleep(seconds, reactor=self.reactor)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Portions Copyright Buildbot Team Members

import threading

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.util.backoff import BackoffTimeoutExceededError
from buildbot.util.queue import ConnectableThreadQueue


class FakeConnection:
    pass


class TestableConnectableThreadQueue(ConnectableThreadQueue):
    def __init__(self, case, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.case = case
        self.create_connection_called_count = 0
        self.close_connection_called_count = 0
        self._test_conn = None

    def create_connection(self):
        self.case.assertTrue(self.connecting)
        self.create_connection_called_count += 1
        self.case.assertIsNone(self._test_conn)
        self._test_conn = FakeConnection()
        return self._test_conn

    def on_close_connection(self, conn):
        self.case.assertIs(conn, self._test_conn)
        self._test_conn = None
        self.close_connection()

    def close_connection(self):
        self.case.assertFalse(self.connecting)
        self._test_conn = None
        self.close_connection_called_count += 1
        super().close_connection()


class TestException(Exception):
    pass


class TestConnectableThreadQueue(unittest.TestCase):

    def setUp(self):
        self.queue = TestableConnectableThreadQueue(self, connect_backoff_start_seconds=0,
                                                    connect_backoff_multiplier=0,
                                                    connect_backoff_max_wait_seconds=0)

    def tearDown(self):
        self.join_queue()

    def join_queue(self, connection_called_count=None):
        self.queue.join(timeout=1)
        if self.queue.is_alive():
            raise AssertionError('Thread is still alive')
        if connection_called_count is not None:
            self.assertEqual(self.queue.create_connection_called_count, connection_called_count)
            self.assertEqual(self.queue.close_connection_called_count, connection_called_count)

    def test_no_work(self):
        self.join_queue(0)

    @defer.inlineCallbacks
    def test_single_item_called(self):
        def work(conn, *args, **kwargs):
            self.assertIs(conn, self.queue.conn)
            self.assertEqual(args, ('arg',))
            self.assertEqual(kwargs, {'kwarg': 'kwvalue'})
            return 'work_result'

        result = yield self.queue.execute_in_thread(work, 'arg', kwarg='kwvalue')
        self.assertEqual(result, 'work_result')

        self.join_queue(1)

    @defer.inlineCallbacks
    def test_single_item_called_exception(self):
        def work(conn):
            raise TestException()

        with self.assertRaises(TestException):
            yield self.queue.execute_in_thread(work)

        self.join_queue(1)

    @defer.inlineCallbacks
    def test_exception_does_not_break_further_work(self):
        def work_exception(conn):
            raise TestException()

        def work_success(conn):
            return 'work_result'

        with self.assertRaises(TestException):
            yield self.queue.execute_in_thread(work_exception)

        result = yield self.queue.execute_in_thread(work_success)
        self.assertEqual(result, 'work_result')

        self.join_queue(1)

    @defer.inlineCallbacks
    def test_single_item_called_disconnect(self):
        def work(conn):
            pass

        yield self.queue.execute_in_thread(work)

        self.queue.close_connection()

        yield self.queue.execute_in_thread(work)

        self.join_queue(2)

    @defer.inlineCallbacks
    def test_many_items_called_in_order(self):
        self.expected_work_index = 0

        def work(conn, work_index):
            self.assertEqual(self.expected_work_index, work_index)
            self.expected_work_index = work_index + 1
            return work_index

        work_deferreds = [self.queue.execute_in_thread(work, i) for i in range(0, 100)]

        for i, d in enumerate(work_deferreds):
            self.assertEqual((yield d), i)

        self.join_queue(1)


class FailingConnectableThreadQueue(ConnectableThreadQueue):
    def __init__(self, case, lock, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.case = case
        self.lock = lock
        self.create_connection_called_count = 0

    def on_close_connection(self, conn):
        raise AssertionError("on_close_connection should not have been called")

    def close_connection(self):
        raise AssertionError("close_connection should not have been called")

    def _drain_queue_with_exception(self, e):
        with self.lock:
            return super()._drain_queue_with_exception(e)


class ThrowingConnectableThreadQueue(FailingConnectableThreadQueue):
    def create_connection(self):
        with self.lock:
            self.create_connection_called_count += 1
            self.case.assertTrue(self.connecting)
            raise TestException()


class NoneReturningConnectableThreadQueue(FailingConnectableThreadQueue):
    def create_connection(self):
        with self.lock:
            self.create_connection_called_count += 1
            self.case.assertTrue(self.connecting)
            return None


class ConnectionErrorTests:

    def setUp(self):
        self.lock = threading.Lock()
        self.queue = self.QueueClass(self, self.lock,
                                     connect_backoff_start_seconds=0.001,
                                     connect_backoff_multiplier=1,
                                     connect_backoff_max_wait_seconds=0.0039)

    def tearDown(self):
        self.queue.join(timeout=1)
        if self.queue.is_alive():
            raise AssertionError('Thread is still alive')

    @defer.inlineCallbacks
    def test_resets_after_reject(self):
        def work(conn):
            raise AssertionError('work should not be executed')

        with self.lock:
            d = self.queue.execute_in_thread(work)

        with self.assertRaises(BackoffTimeoutExceededError):
            yield d

        self.assertEqual(self.queue.create_connection_called_count, 5)

        with self.lock:
            d = self.queue.execute_in_thread(work)

        with self.assertRaises(BackoffTimeoutExceededError):
            yield d

        self.assertEqual(self.queue.create_connection_called_count, 10)
        self.flushLoggedErrors(TestException)

    @defer.inlineCallbacks
    def test_multiple_work_rejected(self):
        def work(conn):
            raise AssertionError('work should not be executed')

        with self.lock:
            d1 = self.queue.execute_in_thread(work)
            d2 = self.queue.execute_in_thread(work)
            d3 = self.queue.execute_in_thread(work)

        with self.assertRaises(BackoffTimeoutExceededError):
            yield d1
        with self.assertRaises(BackoffTimeoutExceededError):
            yield d2
        with self.assertRaises(BackoffTimeoutExceededError):
            yield d3

        self.assertEqual(self.queue.create_connection_called_count, 5)
        self.flushLoggedErrors(TestException)


class TestConnectionErrorThrow(ConnectionErrorTests, unittest.TestCase):
    QueueClass = ThrowingConnectableThreadQueue


class TestConnectionErrorReturnNone(ConnectionErrorTests, unittest.TestCase):
    QueueClass = NoneReturningConnectableThreadQueue



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import textwrap

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.data import logchunks
from buildbot.data import resultspec
from buildbot.test import fakedb
from buildbot.test.util import endpoint


class LogChunkEndpointBase(endpoint.EndpointMixin, unittest.TestCase):

    endpointClass = logchunks.LogChunkEndpoint
    resourceTypeClass = logchunks.LogChunk
    endpointname = "contents"
    log60Lines = ['line zero', 'line 1', 'line TWO', 'line 3', 'line 2**2',
                  'another line', 'yet another line']
    log61Lines = [f'{i:08d}' for i in range(100)]

    def setUp(self):
        self.setUpEndpoint()
        self.db.insertTestData([
            fakedb.Builder(id=77),
            fakedb.Worker(id=13, name='wrk'),
            fakedb.Master(id=88),
            fakedb.Buildset(id=8822),
            fakedb.BuildRequest(id=82, buildsetid=8822),
            fakedb.Build(id=13, builderid=77, masterid=88, workerid=13,
                         buildrequestid=82, number=3),
            fakedb.Step(id=50, buildid=13, number=9, name='make'),
            fakedb.Log(id=60, stepid=50, name='stdio', slug='stdio', type='s',
                       num_lines=7),
            fakedb.LogChunk(logid=60, first_line=0, last_line=1, compressed=0,
                            content=textwrap.dedent("""\
                        line zero
                        line 1""")),
            fakedb.LogChunk(logid=60, first_line=2, last_line=4, compressed=0,
                            content=textwrap.dedent("""\
                        line TWO
                        line 3
                        line 2**2""")),
            fakedb.LogChunk(logid=60, first_line=5, last_line=5, compressed=0,
                            content="another line"),
            fakedb.LogChunk(logid=60, first_line=6, last_line=6, compressed=0,
                            content="yet another line"),
            fakedb.Log(id=61, stepid=50, name='errors', slug='errors',
                       type='t', num_lines=100),
        ] + [
            fakedb.LogChunk(logid=61, first_line=i, last_line=i, compressed=0,
                            content=f"{i:08d}")
            for i in range(100)
        ] + [
            fakedb.Log(id=62, stepid=50, name='notes', slug='notes', type='t',
                       num_lines=0),
            # logid 62 is empty
        ])

    def tearDown(self):
        self.tearDownEndpoint()

    @defer.inlineCallbacks
    def do_test_chunks(self, path, logid, expLines):
        # get the whole thing in one go
        logchunk = yield self.callGet(path)
        self.validateData(logchunk)
        expContent = '\n'.join(expLines) + '\n'
        self.assertEqual(logchunk,
                         {'logid': logid, 'firstline': 0, 'content': expContent})

        # line-by-line
        for i, expLine in enumerate(expLines):
            logchunk = yield self.callGet(path,
                                          resultSpec=resultspec.ResultSpec(offset=i, limit=1))
            self.validateData(logchunk)
            self.assertEqual(logchunk,
                             {'logid': logid, 'firstline': i, 'content': expLine + '\n'})

        # half and half
        mid = int(len(expLines) / 2)
        for f, length in (0, mid), (mid, len(expLines) - 1):
            result_spec = resultspec.ResultSpec(offset=f, limit=length - f + 1)
            logchunk = yield self.callGet(path, resultSpec=result_spec)
            self.validateData(logchunk)
            expContent = '\n'.join(expLines[f:length + 1]) + '\n'
            self.assertEqual(logchunk,
                             {'logid': logid, 'firstline': f, 'content': expContent})

        # truncated at EOF
        f, length = len(expLines) - 2, len(expLines) + 10
        result_spec = resultspec.ResultSpec(offset=f, limit=length - f + 1)
        logchunk = yield self.callGet(path, resultSpec=result_spec)
        self.validateData(logchunk)
        expContent = '\n'.join(expLines[-2:]) + '\n'
        self.assertEqual(logchunk,
                         {'logid': logid, 'firstline': f, 'content': expContent})

        # some illegal stuff
        self.assertEqual(
            (yield self.callGet(path,
                                resultSpec=resultspec.ResultSpec(offset=-1))), None)
        self.assertEqual(
            (yield self.callGet(path,
                                resultSpec=resultspec.ResultSpec(offset=10, limit=-1))),
            None)

    def test_get_logid_60(self):
        return self.do_test_chunks(('logs', 60, self.endpointname), 60,
                                   self.log60Lines)

    def test_get_logid_61(self):
        return self.do_test_chunks(('logs', 61, self.endpointname), 61,
                                   self.log61Lines)


class LogChunkEndpoint(LogChunkEndpointBase):

    @defer.inlineCallbacks
    def test_get_missing(self):
        logchunk = yield self.callGet(('logs', 99, self.endpointname))
        self.assertEqual(logchunk, None)

    @defer.inlineCallbacks
    def test_get_empty(self):
        logchunk = yield self.callGet(('logs', 62, self.endpointname))
        self.validateData(logchunk)
        self.assertEqual(logchunk['content'], '')

    @defer.inlineCallbacks
    def test_get_by_stepid(self):
        logchunk = yield self.callGet(
            ('steps', 50, 'logs', 'errors', self.endpointname))
        self.validateData(logchunk)
        self.assertEqual(logchunk['logid'], 61)

    @defer.inlineCallbacks
    def test_get_by_buildid(self):
        logchunk = yield self.callGet(
            ('builds', 13, 'steps', 9, 'logs', 'stdio', self.endpointname))
        self.validateData(logchunk)
        self.assertEqual(logchunk['logid'], 60)

    @defer.inlineCallbacks
    def test_get_by_builder(self):
        logchunk = yield self.callGet(
            ('builders', 77, 'builds', 3, 'steps', 9,
             'logs', 'errors', self.endpointname))
        self.validateData(logchunk)
        self.assertEqual(logchunk['logid'], 61)

    @defer.inlineCallbacks
    def test_get_by_builder_step_name(self):
        logchunk = yield self.callGet(
            ('builders', 77, 'builds', 3, 'steps', 'make',
             'logs', 'errors', self.endpointname))
        self.validateData(logchunk)
        self.assertEqual(logchunk['logid'], 61)


class RawLogChunkEndpoint(LogChunkEndpointBase):

    endpointClass = logchunks.RawLogChunkEndpoint
    endpointname = "raw"

    def validateData(self, data):
        self.assertIsInstance(data['raw'], str)
        self.assertIsInstance(data['mime-type'], str)
        self.assertIsInstance(data['filename'], str)

    @defer.inlineCallbacks
    def do_test_chunks(self, path, logid, expLines):
        # get the whole thing in one go
        logchunk = yield self.callGet(path)
        self.validateData(logchunk)
        if logid == 60:
            expContent = '\n'.join([line[1:] for line in expLines])
            expFilename = "stdio"
        else:
            expContent = '\n'.join(expLines) + '\n'
            expFilename = "errors"

        self.assertEqual(logchunk,
                         {'filename': expFilename, 'mime-type': "text/plain", 'raw': expContent})



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import calendar
import datetime

import jwt

import mock

from twisted.cred import strcred
from twisted.cred.checkers import InMemoryUsernamePasswordDatabaseDontUse
from twisted.internet import defer
from twisted.trial import unittest
from twisted.web._auth.wrapper import HTTPAuthSessionWrapper
from twisted.web.server import Request

from buildbot.test.reactor import TestReactorMixin
from buildbot.test.unit.www import test_hooks_base
from buildbot.test.util import www
from buildbot.www import auth
from buildbot.www import change_hook
from buildbot.www import resource
from buildbot.www import rest
from buildbot.www import service


class FakeChannel:
    transport = None

    def isSecure(self):
        return False

    def getPeer(self):
        return None

    def getHost(self):
        return None


class NeedsReconfigResource(resource.Resource):

    needsReconfig = True
    reconfigs = 0

    def reconfigResource(self, new_config):
        NeedsReconfigResource.reconfigs += 1


class Test(TestReactorMixin, www.WwwTestMixin, unittest.TestCase):

    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.master = self.make_master(url='h:/a/b/')
        self.svc = self.master.www = service.WWWService()
        yield self.svc.setServiceParent(self.master)

    def makeConfig(self, **kwargs):
        w = dict(port=None, auth=auth.NoAuth(), logfileName='l')
        w.update(kwargs)
        new_config = mock.Mock()
        new_config.www = w
        new_config.buildbotURL = 'h:/'
        self.master.config = new_config
        return new_config

    @defer.inlineCallbacks
    def test_reconfigService_no_port(self):
        new_config = self.makeConfig()
        yield self.svc.reconfigServiceWithBuildbotConfig(new_config)

        self.assertEqual(self.svc.site, None)

    @defer.inlineCallbacks
    def test_reconfigService_reconfigResources(self):
        new_config = self.makeConfig(port=8080)
        self.patch(rest, 'RestRootResource', NeedsReconfigResource)
        NeedsReconfigResource.reconfigs = 0

        # first time, reconfigResource gets called along with setupSite
        yield self.svc.reconfigServiceWithBuildbotConfig(new_config)
        self.assertEqual(NeedsReconfigResource.reconfigs, 1)

        # and the next time, setupSite isn't called, but reconfigResource is
        yield self.svc.reconfigServiceWithBuildbotConfig(new_config)
        self.assertEqual(NeedsReconfigResource.reconfigs, 2)

    @defer.inlineCallbacks
    def test_reconfigService_port(self):
        new_config = self.makeConfig(port=20)
        yield self.svc.reconfigServiceWithBuildbotConfig(new_config)

        self.assertNotEqual(self.svc.site, None)
        self.assertNotEqual(self.svc.port_service, None)
        self.assertEqual(self.svc.port, 20)

    @defer.inlineCallbacks
    def test_reconfigService_expiration_time(self):
        new_config = self.makeConfig(port=80, cookie_expiration_time=datetime.timedelta(minutes=1))
        yield self.svc.reconfigServiceWithBuildbotConfig(new_config)

        self.assertNotEqual(self.svc.site, None)
        self.assertNotEqual(self.svc.port_service, None)
        self.assertEqual(service.BuildbotSession.expDelay, datetime.timedelta(minutes=1))

    @defer.inlineCallbacks
    def test_reconfigService_port_changes(self):
        new_config = self.makeConfig(port=20)
        yield self.svc.reconfigServiceWithBuildbotConfig(new_config)

        newer_config = self.makeConfig(port=999)
        yield self.svc.reconfigServiceWithBuildbotConfig(newer_config)

        self.assertNotEqual(self.svc.site, None)
        self.assertNotEqual(self.svc.port_service, None)
        self.assertEqual(self.svc.port, 999)

    @defer.inlineCallbacks
    def test_reconfigService_port_changes_to_none(self):
        new_config = self.makeConfig(port=20)
        yield self.svc.reconfigServiceWithBuildbotConfig(new_config)

        newer_config = self.makeConfig()
        yield self.svc.reconfigServiceWithBuildbotConfig(newer_config)

        # (note the site sticks around)
        self.assertEqual(self.svc.port_service, None)
        self.assertEqual(self.svc.port, None)

    def test_setupSite(self):
        self.svc.setupSite(self.makeConfig())
        site = self.svc.site

        # check that it has the right kind of resources attached to its
        # root
        root = site.resource
        req = mock.Mock()
        self.assertIsInstance(root.getChildWithDefault(b'api', req),
                              rest.RestRootResource)

    def test_setupSiteWithProtectedHook(self):
        checker = InMemoryUsernamePasswordDatabaseDontUse()
        checker.addUser("guest", "password")

        self.svc.setupSite(self.makeConfig(
            change_hook_dialects={'base': True},
            change_hook_auth=[checker]))
        site = self.svc.site

        # check that it has the right kind of resources attached to its
        # root
        root = site.resource
        req = mock.Mock()
        self.assertIsInstance(root.getChildWithDefault(b'change_hook', req),
                              HTTPAuthSessionWrapper)

    @defer.inlineCallbacks
    def test_setupSiteWithHook(self):
        new_config = self.makeConfig(
            change_hook_dialects={'base': True})
        self.svc.setupSite(new_config)
        site = self.svc.site

        # check that it has the right kind of resources attached to its
        # root
        root = site.resource
        req = mock.Mock()
        ep = root.getChildWithDefault(b'change_hook', req)
        self.assertIsInstance(ep,
                              change_hook.ChangeHookResource)

        # not yet configured
        self.assertEqual(ep.dialects, {})

        yield self.svc.reconfigServiceWithBuildbotConfig(new_config)

        # now configured
        self.assertEqual(ep.dialects, {'base': True})

        rsrc = self.svc.site.resource.getChildWithDefault(b'change_hook', mock.Mock())
        path = b'/change_hook/base'
        request = test_hooks_base._prepare_request({})
        self.master.data.updates.addChange = mock.Mock()
        yield self.render_resource(rsrc, path, request=request)
        self.master.data.updates.addChange.assert_called()

    @defer.inlineCallbacks
    def test_setupSiteWithHookAndAuth(self):
        fn = self.mktemp()
        with open(fn, 'w', encoding='utf-8') as f:
            f.write("user:pass")
        new_config = self.makeConfig(
            port=8080,
            plugins={},
            change_hook_dialects={'base': True},
            change_hook_auth=[strcred.makeChecker("file:" + fn)])
        self.svc.setupSite(new_config)

        yield self.svc.reconfigServiceWithBuildbotConfig(new_config)
        rsrc = self.svc.site.resource.getChildWithDefault(b'', mock.Mock())

        res = yield self.render_resource(rsrc, b'')
        self.assertIn(b'{"type": "file"}', res)

        rsrc = self.svc.site.resource.getChildWithDefault(
            b'change_hook', mock.Mock())
        res = yield self.render_resource(rsrc, b'/change_hook/base')
        # as UnauthorizedResource is in private namespace, we cannot use
        # assertIsInstance :-(
        self.assertIn('UnauthorizedResource', repr(res))


class TestBuildbotSite(unittest.SynchronousTestCase):
    SECRET = 'secret'

    def setUp(self):
        self.site = service.BuildbotSite(None, "logs", 0, 0)
        self.site.setSessionSecret(self.SECRET)

    def test_getSession_from_bad_jwt(self):
        """ if the cookie is bad (maybe from previous version of buildbot),
            then we should raise KeyError for consumption by caller,
            and log the JWT error
        """
        with self.assertRaises(KeyError):
            self.site.getSession("xxx")
        self.flushLoggedErrors(jwt.exceptions.DecodeError)

    def test_getSession_from_correct_jwt(self):
        payload = {'user_info': {'some': 'payload'}}
        uid = jwt.encode(payload, self.SECRET, algorithm=service.SESSION_SECRET_ALGORITHM)
        session = self.site.getSession(uid)
        self.assertEqual(session.user_info, {'some': 'payload'})

    def test_getSession_from_expired_jwt(self):
        # expired one week ago
        exp = datetime.datetime.utcnow() - datetime.timedelta(weeks=1)
        exp = calendar.timegm(datetime.datetime.timetuple(exp))
        payload = {'user_info': {'some': 'payload'}, 'exp': exp}
        uid = jwt.encode(payload, self.SECRET, algorithm=service.SESSION_SECRET_ALGORITHM)
        with self.assertRaises(KeyError):
            self.site.getSession(uid)

    def test_getSession_with_no_user_info(self):
        payload = {'foo': 'bar'}
        uid = jwt.encode(payload, self.SECRET, algorithm=service.SESSION_SECRET_ALGORITHM)
        with self.assertRaises(KeyError):
            self.site.getSession(uid)

    def test_makeSession(self):
        session = self.site.makeSession()
        self.assertEqual(session.user_info, {'anonymous': True})

    def test_updateSession(self):
        session = self.site.makeSession()
        request = Request(FakeChannel(), False)
        request.sitepath = [b"bb"]
        session.updateSession(request)
        self.assertEqual(len(request.cookies), 1)
        _, value = request.cookies[0].split(b";")[0].split(b"=")
        decoded = jwt.decode(value, self.SECRET,
                             algorithms=[service.SESSION_SECRET_ALGORITHM])
        self.assertEqual(decoded['user_info'], {'anonymous': True})
        self.assertIn('exp', decoded)

    def test_absentServerHeader(self):
        request = Request(FakeChannel(), False)
        self.assertEqual(request.responseHeaders.hasHeader('Server'), False)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


import base64

import sqlalchemy as sa

from twisted.internet import defer
from twisted.python import log

from buildbot.db import base
from buildbot.util import bytes2unicode
from buildbot.util import epoch2datetime
from buildbot.util import unicode2bytes


class SsDict(dict):
    pass


class SsList(list):
    pass


class SourceStampsConnectorComponent(base.DBConnectorComponent):
    # Documentation is in developer/database.rst

    @defer.inlineCallbacks
    def findSourceStampId(self, branch=None, revision=None, repository=None,
                          project=None, codebase=None, patch_body=None,
                          patch_level=None, patch_author=None,
                          patch_comment=None, patch_subdir=None):
        sourcestampid, _ = yield self.findOrCreateId(
            branch, revision, repository, project, codebase, patch_body,
            patch_level, patch_author, patch_comment, patch_subdir)
        return sourcestampid

    @defer.inlineCallbacks
    def findOrCreateId(self, branch=None, revision=None, repository=None,
                       project=None, codebase=None, patch_body=None,
                       patch_level=None, patch_author=None,
                       patch_comment=None, patch_subdir=None):
        tbl = self.db.model.sourcestamps

        assert codebase is not None, "codebase cannot be None"
        assert project is not None, "project cannot be None"
        assert repository is not None, "repository cannot be None"
        self.checkLength(tbl.c.branch, branch)
        self.checkLength(tbl.c.revision, revision)
        self.checkLength(tbl.c.repository, repository)
        self.checkLength(tbl.c.project, project)

        # get a patchid, if we have a patch
        def thd(conn):
            patchid = None
            if patch_body:
                patch_body_bytes = unicode2bytes(patch_body)
                patch_base64_bytes = base64.b64encode(patch_body_bytes)
                ins = self.db.model.patches.insert()
                r = conn.execute(ins, dict(
                    patchlevel=patch_level,
                    patch_base64=bytes2unicode(patch_base64_bytes),
                    patch_author=patch_author,
                    patch_comment=patch_comment,
                    subdir=patch_subdir))
                patchid = r.inserted_primary_key[0]
            return patchid
        patchid = yield self.db.pool.do(thd)

        ss_hash = self.hashColumns(branch, revision, repository, project,
                                   codebase, patchid)
        sourcestampid, found = yield self.findOrCreateSomethingId(
            tbl=tbl,
            whereclause=tbl.c.ss_hash == ss_hash,
            insert_values={
                'branch': branch,
                'revision': revision,
                'repository': repository,
                'codebase': codebase,
                'project': project,
                'patchid': patchid,
                'ss_hash': ss_hash,
                'created_at': int(self.master.reactor.seconds()),
            })
        return sourcestampid, found

    # returns a Deferred that returns a value
    @base.cached("ssdicts")
    def getSourceStamp(self, ssid):
        def thd(conn):
            tbl = self.db.model.sourcestamps
            q = tbl.select(whereclause=(tbl.c.id == ssid))
            res = conn.execute(q)
            row = res.fetchone()
            if not row:
                return None
            ssdict = self._rowToSsdict_thd(conn, row)
            res.close()
            return ssdict
        return self.db.pool.do(thd)

    # returns a Deferred that returns a value
    def getSourceStampsForBuild(self, buildid):
        assert buildid > 0

        def thd(conn):
            # Get SourceStamps for the build
            builds_tbl = self.db.model.builds
            reqs_tbl = self.db.model.buildrequests
            bsets_tbl = self.db.model.buildsets
            bsss_tbl = self.db.model.buildset_sourcestamps
            sstamps_tbl = self.db.model.sourcestamps

            from_clause = builds_tbl.join(reqs_tbl,
                                          builds_tbl.c.buildrequestid == reqs_tbl.c.id)
            from_clause = from_clause.join(bsets_tbl,
                                           reqs_tbl.c.buildsetid == bsets_tbl.c.id)
            from_clause = from_clause.join(bsss_tbl,
                                           bsets_tbl.c.id == bsss_tbl.c.buildsetid)
            from_clause = from_clause.join(sstamps_tbl,
                                           bsss_tbl.c.sourcestampid == sstamps_tbl.c.id)

            q = sa.select([sstamps_tbl]).select_from(
                from_clause).where(builds_tbl.c.id == buildid)
            res = conn.execute(q)
            return [self._rowToSsdict_thd(conn, row)
                    for row in res.fetchall()]

        return self.db.pool.do(thd)

    # returns a Deferred that returns a value
    def getSourceStamps(self):
        def thd(conn):
            tbl = self.db.model.sourcestamps
            q = tbl.select()
            res = conn.execute(q)
            return [self._rowToSsdict_thd(conn, row)
                    for row in res.fetchall()]
        return self.db.pool.do(thd)

    def _rowToSsdict_thd(self, conn, row):
        ssid = row.id
        ssdict = SsDict(ssid=ssid, branch=row.branch,
                        revision=row.revision, patchid=None, patch_body=None,
                        patch_level=None, patch_author=None, patch_comment=None,
                        patch_subdir=None, repository=row.repository,
                        codebase=row.codebase, project=row.project,
                        created_at=epoch2datetime(row.created_at))
        patchid = row.patchid

        # fetch the patch, if necessary
        if patchid is not None:
            tbl = self.db.model.patches
            q = tbl.select(whereclause=(tbl.c.id == patchid))
            res = conn.execute(q)
            row = res.fetchone()
            if row:
                # note the subtle renaming here
                ssdict['patchid'] = patchid
                ssdict['patch_level'] = row.patchlevel
                ssdict['patch_subdir'] = row.subdir
                ssdict['patch_author'] = row.patch_author
                ssdict['patch_comment'] = row.patch_comment
                ssdict['patch_body'] = base64.b64decode(row.patch_base64)
            else:
                log.msg(f'patchid {patchid}, referenced from ssid {ssid}, not found')
            res.close()
        return ssdict
